from typing import Any, Dict, Type, Union, TYPE_CHECKING, Optional

from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.destination.capabilities import DataTypeMapper
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.normalizers import NamingConvention

from dlt.destinations.impl.sqlalchemy.configuration import (
    SqlalchemyCredentials,
    SqlalchemyClientConfiguration,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.sqlalchemy.sqlalchemy_job_client import SqlalchemyJobClient
    from sqlalchemy.engine import Engine
else:
    Engine = Any


class sqlalchemy(Destination[SqlalchemyClientConfiguration, "SqlalchemyJobClient"]):
    spec = SqlalchemyClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        # lazy import to avoid sqlalchemy dep
        SqlalchemyTypeMapper: Type[DataTypeMapper]

        try:
            from dlt.destinations.impl.sqlalchemy.type_mapper import SqlalchemyTypeMapper
        except ModuleNotFoundError:
            # assign mock type mapper if no sqlalchemy
            from dlt.common.destination.capabilities import (
                UnsupportedTypeMapper as SqlalchemyTypeMapper,
            )

        # https://www.sqlalchemyql.org/docs/current/limits.html
        caps = DestinationCapabilitiesContext.generic_capabilities()
        caps.preferred_loader_file_format = "typed-jsonl"
        caps.supported_loader_file_formats = ["typed-jsonl", "parquet", "model"]
        caps.preferred_staging_file_format = None
        caps.supported_staging_file_formats = []
        caps.has_case_sensitive_identifiers = True
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)
        caps.max_identifier_length = 63
        caps.max_column_identifier_length = 63
        caps.max_query_length = 32 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 1024 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = True
        caps.max_query_parameters = 20_0000
        caps.max_rows_per_insert = 10_000  # Set a default to avoid OOM on large datasets
        # Multiple concatenated statements are not supported by all engines, so leave them off by default
        caps.supports_multiple_statements = False
        caps.type_mapper = SqlalchemyTypeMapper
        caps.supported_replace_strategies = ["truncate-and-insert", "insert-from-staging"]
        caps.supported_merge_strategies = ["delete-insert", "scd2"]

        return caps

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: SqlalchemyClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        dialect_class = config.get_dialect_class()
        if dialect_class is None:
            return super(sqlalchemy, cls).adjust_capabilities(caps, config, naming)

        dialect = dialect_class()
        backend_name = config.get_backend_name()

        caps.max_identifier_length = dialect.max_identifier_length
        caps.max_column_identifier_length = dialect.max_identifier_length
        caps.supports_native_boolean = dialect.supports_native_boolean

        from dlt.destinations.impl.sqlalchemy.dialect import (
            get_dialect_capabilities,
            DialectCapabilities,
        )

        dialect_caps = (
            get_dialect_capabilities(backend_name)
            or get_dialect_capabilities(dialect.name)
            or DialectCapabilities(backend_name)
        )

        caps.sqlglot_dialect = dialect_caps.sqlglot_dialect  # type: ignore[assignment]
        caps.type_mapper = dialect_caps.type_mapper_class()
        caps.dialect_capabilities = dialect_caps
        dialect_caps.adjust_capabilities(caps, dialect)

        if dialect.requires_name_normalize:  # type: ignore[attr-defined]
            caps.has_case_sensitive_identifiers = False
            caps.casefold_identifier = str.lower

        return super(sqlalchemy, cls).adjust_capabilities(caps, config, naming)

    @property
    def client_class(self) -> Type["SqlalchemyJobClient"]:
        from dlt.destinations.impl.sqlalchemy.sqlalchemy_job_client import SqlalchemyJobClient

        return SqlalchemyJobClient

    def __init__(
        self,
        credentials: Union[SqlalchemyCredentials, Dict[str, Any], str, Engine] = None,
        create_unique_indexes: bool = False,
        create_primary_keys: bool = False,
        destination_name: str = None,
        environment: str = None,
        engine_kwargs: Dict[str, Any] = None,
        engine_args: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Sqlalchemy destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[SqlalchemyCredentials, Dict[str, Any], str, Engine], optional): Credentials to connect to the sqlalchemy database. Can be an instance of
                `SqlalchemyCredentials` or a connection string in the format `mysql://user:password@host:port/database`. Defaults to None.
            create_unique_indexes (bool, optional): Whether UNIQUE constraints should be created. Defaults to False.
            create_primary_keys (bool, optional): Whether PRIMARY KEY constraints should be created. Defaults to False.
            destination_name (str, optional): The name of the destination. Defaults to None.
            environment (str, optional): The environment to use. Defaults to None.
            engine_kwargs (Dict[str, Any], optional): Additional keyword arguments passed to `sqlalchemy.create_engine`. Defaults to None.
            engine_args (Dict[str, Any], optional): Deprecated. Use engine_kwargs instead. Defaults to None.
            **kwargs (Any): Additional arguments passed to the destination.
        """
        super().__init__(
            credentials=credentials,
            create_unique_indexes=create_unique_indexes,
            create_primary_keys=create_primary_keys,
            destination_name=destination_name,
            environment=environment,
            engine_kwargs=engine_kwargs,
            engine_args=engine_args,
            **kwargs,
        )


sqlalchemy.register()



from typing import Sequence, Type, cast, overload, Optional
from typing_extensions import TypeVar

from dlt.common.configuration.specs import known_sections
from dlt.common.schema import Schema
from dlt.common.schema.typing import (
    TColumnSchema,
    TTableFormat,
    TWriteDispositionConfig,
    TSchemaContract,
)

from dlt.common.typing import TSecretStrValue, Any
from dlt.common.configuration import with_config
from dlt.common.configuration.container import Container
from dlt.common.configuration.inject import get_orig_args, last_config
from dlt.common.destination import TLoaderFileFormat, Destination, TDestinationReferenceArg
from dlt.common.pipeline import LoadInfo, PipelineContext, get_dlt_pipelines_dir, TRefreshMode

# from dlt.common.runtime import apply_runtime_config, init_telemetry
from dlt.pipeline.exceptions import CannotRestorePipelineException

from dlt.pipeline.configuration import PipelineConfiguration, ensure_correct_pipeline_kwargs
from dlt.pipeline.pipeline import Pipeline
from dlt.pipeline.progress import _from_name as collector_from_name, TCollectorArg, _NULL_COLLECTOR
from dlt.pipeline.warnings import full_refresh_argument_deprecated

TPipeline = TypeVar("TPipeline", bound=Pipeline, default=Pipeline)


@overload
def pipeline(
    pipeline_name: str = None,
    pipelines_dir: str = None,
    pipeline_salt: TSecretStrValue = None,
    destination: TDestinationReferenceArg = None,
    staging: TDestinationReferenceArg = None,
    dataset_name: str = None,
    import_schema_path: str = None,
    export_schema_path: str = None,
    full_refresh: bool = None,
    dev_mode: bool = False,
    refresh: TRefreshMode = None,
    progress: TCollectorArg = _NULL_COLLECTOR,
    _impl_cls: Type[TPipeline] = Pipeline,  # type: ignore[assignment]
) -> TPipeline:
    """Creates a new instance of `dlt` pipeline, which moves the data from the source ie. a REST API to a destination ie. database or a data lake.

    #### Note:
    The `pipeline` functions allows you to pass the destination name to which the data should be loaded, the name of the dataset and several other options that govern loading of the data.
    The created `Pipeline` object lets you load the data from any source with `run` method or to have more granular control over the loading process with `extract`, `normalize` and `load` methods.

    Please refer to the following doc pages
    - Write your first pipeline walkthrough: https://dlthub.com/docs/walkthroughs/create-a-pipeline
    - Pipeline architecture and data loading steps: https://dlthub.com/docs/reference
    - List of supported destinations: https://dlthub.com/docs/dlt-ecosystem/destinations

    Args:
        pipeline_name (str, optional): A name of the pipeline that will be used to identify it in monitoring events and to restore its state and data schemas on subsequent runs.
            Defaults to the file name of pipeline script with `dlt_` prefix added.

        pipelines_dir (str, optional): A working directory in which pipeline state and temporary files will be stored. Defaults to user home directory: `~/dlt/pipelines/`.

        pipeline_salt (TSecretStrValue, optional): A random value used for deterministic hashing during data anonymization. Defaults to a value derived from the pipeline name.
            Default value should not be used for any cryptographic purposes.

        destination (TDestinationReferenceArg, optional): A name of the destination to which dlt will load the data, or a destination module imported from `dlt.destination`.
            May also be provided to `run` method of the `pipeline`.

        staging (TDestinationReferenceArg, optional): A name of the destination where dlt will stage the data before final loading, or a destination module imported from `dlt.destination`.
            May also be provided to `run` method of the `pipeline`.

        dataset_name (str, optional): A name of the dataset to which the data will be loaded. A dataset is a logical group of tables ie. `schema` in relational databases or folder grouping many files.
            May also be provided later to the `run` or `load` methods of the `Pipeline`. If not provided at all then defaults to the `pipeline_name`

        import_schema_path (str, optional): A path from which the schema `yaml` file will be imported on each pipeline run. Defaults to None which disables importing.

        export_schema_path (str, optional): A path where the schema `yaml` file will be exported after every schema change. Defaults to None which disables exporting.

        full_refresh (bool, optional): Deprecated use dev_mode instead.

        dev_mode (bool, optional): When set to True, each instance of the pipeline with the `pipeline_name` starts from scratch when run and loads the data to a separate dataset.
            The datasets are identified by `dataset_name_` + datetime suffix. Use this setting whenever you experiment with your data to be sure you start fresh on each run. Defaults to False.

        refresh (TRefreshMode, optional): Fully or partially reset sources during pipeline run. When set here the refresh is applied on each run of the pipeline.
            To apply refresh only once you can pass it to `pipeline.run` or `extract` instead. The following refresh modes are supported:
            * `drop_sources`: Drop tables and source and resource state for all sources currently being processed in `run` or `extract` methods of the pipeline. (Note: schema history is erased)
            * `drop_resources`: Drop tables and resource state for all resources being processed. Source level state is not modified. (Note: schema history is erased)
            * `drop_data`: Wipe all data and resource state for all resources being processed. Schema is not modified.

        progress(TCollectorArg): A progress monitor that shows progress bars, console or log messages with current information on sources, resources, data items etc. processed in
            `extract`, `normalize` and `load` stage. Pass a string with a collector name or configure your own by choosing from `dlt.progress` module.
            We support most of the progress libraries: try passing `tqdm`, `enlighten` or `alive_progress` or `log` to write to console/log.

        _impl_cls (Type[TPipeline], optional): A class of the pipeline to use. Defaults to `Pipeline`.

    Returns:
        TPipeline: An instance of `Pipeline` class or a subclass. Please check the documentation of `run` method for information on what to do with it.
    """


@overload
def pipeline() -> Pipeline:  # type: ignore
    """When called without any arguments, returns the recently created `Pipeline` instance.
    If not found, it creates a new instance with all the pipeline options set to defaults."""


@with_config(
    spec=PipelineConfiguration,
    sections=(known_sections.PIPELINES,),
    section_arg_name="pipeline_name",
)
def pipeline(
    pipeline_name: str = None,
    pipelines_dir: str = None,
    pipeline_salt: TSecretStrValue = None,
    destination: TDestinationReferenceArg = None,
    staging: TDestinationReferenceArg = None,
    dataset_name: str = None,
    import_schema_path: str = None,
    export_schema_path: str = None,
    full_refresh: Optional[bool] = None,
    dev_mode: bool = False,
    refresh: Optional[TRefreshMode] = None,
    progress: TCollectorArg = _NULL_COLLECTOR,
    _impl_cls: Type[TPipeline] = Pipeline,  # type: ignore[assignment]
    **injection_kwargs: Any,
) -> TPipeline:
    ensure_correct_pipeline_kwargs(pipeline, **injection_kwargs)
    # call without arguments returns current pipeline
    orig_args = get_orig_args(**injection_kwargs)  # original (*args, **kwargs)
    # is any of the arguments different from defaults
    has_arguments = bool(orig_args[0]) or any(orig_args[1].values())

    full_refresh_argument_deprecated("pipeline", full_refresh)

    if not has_arguments:
        context = Container()[PipelineContext]
        # if pipeline instance is already active then return it, otherwise create a new one
        if context.is_active():
            return cast(TPipeline, context.pipeline())
        else:
            pass

    # if working_dir not provided use temp folder
    if not pipelines_dir:
        pipelines_dir = get_dlt_pipelines_dir()

    destination = Destination.from_reference(
        destination or injection_kwargs["destination_type"],
        destination_name=injection_kwargs["destination_name"],
    )
    staging = Destination.from_reference(
        staging or injection_kwargs.get("staging_type", None),
        destination_name=injection_kwargs.get("staging_name", None),
    )

    progress = collector_from_name(progress)
    # create new pipeline instance
    p = _impl_cls(
        pipeline_name,
        pipelines_dir,
        pipeline_salt,
        destination,
        staging,
        dataset_name,
        import_schema_path,
        export_schema_path,
        full_refresh if full_refresh is not None else dev_mode,
        progress,
        False,
        last_config(**injection_kwargs),
        refresh=refresh,
    )
    # set it as current pipeline
    p.activate()
    return p


@with_config(
    spec=PipelineConfiguration,
    sections=(known_sections.PIPELINES,),
    section_arg_name="pipeline_name",
)
def attach(
    pipeline_name: str = None,
    pipelines_dir: str = None,
    pipeline_salt: TSecretStrValue = None,
    destination: TDestinationReferenceArg = None,
    staging: TDestinationReferenceArg = None,
    progress: TCollectorArg = _NULL_COLLECTOR,
    dataset_name: str = None,
    **injection_kwargs: Any,
) -> Pipeline:
    """Attaches to the working folder of `pipeline_name` in `pipelines_dir` or in default directory.
    Pre-configured `destination` and `staging` factories may be provided. If not present, default factories are created from pipeline state.

    If no local pipeline state is found, dlt will attempt to restore the pipeline from the provided destination and dataset.
    """
    ensure_correct_pipeline_kwargs(attach, **injection_kwargs)

    # if working_dir not provided use temp folder
    if not pipelines_dir:
        pipelines_dir = get_dlt_pipelines_dir()
    progress = collector_from_name(progress)
    destination = Destination.from_reference(
        destination or injection_kwargs["destination_type"],
        destination_name=injection_kwargs["destination_name"],
    )
    staging = Destination.from_reference(
        staging or injection_kwargs.get("staging_type", None),
        destination_name=injection_kwargs.get("staging_name", None),
    )
    # create new pipeline instance
    try:
        p = Pipeline(
            pipeline_name,
            pipelines_dir,
            pipeline_salt,
            destination,
            staging,
            dataset_name,
            None,
            None,
            False,  # always False as dev_mode so we do not wipe the working folder
            progress,
            True,
            last_config(**injection_kwargs),
        )
        # set it as current pipeline
        p.activate()
        return p
    except CannotRestorePipelineException:
        if not pipeline_name:
            raise

        # we can try to sync a pipeline with the given name
        p = pipeline(
            pipeline_name,
            pipelines_dir,
            destination=destination,
            staging=staging,
            dataset_name=dataset_name,
        )

        # we cannot restore the pipeline if the destination is not provided
        # or found in the env, wipe the working folder and re-raise
        if not p.destination:
            p._wipe_working_folder()
            raise CannotRestorePipelineException(
                pipeline_name,
                p.working_dir,
                f"the pipeline was not found in {p.working_dir} found and no destination was"
                " provided to restore from.",
            )
        p.sync_destination()

        # if remote state was not found,
        # wipe the working folder and re-raise
        if p.first_run:
            p._wipe_working_folder()
            raise CannotRestorePipelineException(
                pipeline_name,
                p.working_dir,
                f"the pipeline was not found in {p.working_dir} found and provided destination and"
                " dataset do not contain state for this pipeline.",
            )
        return p


def run(
    data: Any,
    *,
    destination: TDestinationReferenceArg = None,
    staging: TDestinationReferenceArg = None,
    dataset_name: str = None,
    table_name: str = None,
    write_disposition: TWriteDispositionConfig = None,
    columns: Sequence[TColumnSchema] = None,
    schema: Schema = None,
    loader_file_format: TLoaderFileFormat = None,
    table_format: TTableFormat = None,
    schema_contract: TSchemaContract = None,
    refresh: TRefreshMode = None,
) -> LoadInfo:
    """Loads the data in `data` argument into the destination specified in `destination` and dataset specified in `dataset_name`.

    #### Note:
    This method will `extract` the data from the `data` argument, infer the schema, `normalize` the data into a load package (ie. jsonl or PARQUET files representing tables) and then `load` such packages into the `destination`.

    The data may be supplied in several forms:
    * a `list` or `Iterable` of any JSON-serializable objects ie. `dlt.run([1, 2, 3], table_name="numbers")`
    * any `Iterator` or a function that yield (`Generator`) ie. `dlt.run(range(1, 10), table_name="range")`
    * a function or a list of functions decorated with @dlt.resource ie. `dlt.run([chess_players(title="GM"), chess_games()])`
    * a function or a list of functions decorated with @dlt.source.

    Please note that `dlt` deals with `bytes`, `datetime`, `decimal` and `uuid` objects so you are free to load binary data or documents containing dates.

    #### Execution:
    The `run` method will first use `sync_destination` method to synchronize pipeline state and schemas with the destination. You can disable this behavior with `restore_from_destination` configuration option.
    Next it will make sure that data from the previous is fully processed. If not, `run` method normalizes and loads pending data items.
    Only then the new data from `data` argument is extracted, normalized and loaded.

    Args:
        data (Any): Data to be loaded to destination

        destination (TDestinationReferenceArg, optional): A name of the destination to which dlt will load the data, or a destination module imported from `dlt.destination`.
            If not provided, the value passed to `dlt.pipeline` will be used.

        staging (TDestinationReferenceArg, optional): A name of the destination where dlt will stage the data before final loading, or a destination module imported from `dlt.destination`.
            May also be provided to `run` method of the `pipeline`.

        dataset_name (str, optional): A name of the dataset to which the data will be loaded. A dataset is a logical group of tables ie. `schema` in relational databases or folder grouping many files.
            If not provided, the value passed to `dlt.pipeline` will be used. If not provided at all then defaults to the `pipeline_name`

        table_name (str, optional): The name of the table to which the data should be loaded within the `dataset`. This argument is required for a `data` that is a list/Iterable or Iterator without `__name__` attribute.
            The behavior of this argument depends on the type of the `data`:
            * generator functions: the function name is used as table name, `table_name` overrides this default
            * `@dlt.resource`: resource contains the full table schema and that includes the table name. `table_name` will override this property. Use with care!
            * `@dlt.source`: source contains several resources each with a table schema. `table_name` will override all table names within the source and load the data into single table.

        write_disposition (TWriteDispositionConfig, optional): Controls how to write data to a table. Accepts a shorthand string literal or configuration dictionary.
            Allowed shorthand string literals: `append` will always add new data at the end of the table. `replace` will replace existing data with new data. `skip` will prevent data from loading. "merge" will deduplicate and merge data based on "primary_key" and "merge_key" hints. Defaults to "append".
            Write behaviour can be further customized through a configuration dictionary. For example, to obtain an SCD2 table provide `write_disposition={"disposition": "merge", "strategy": "scd2"}`.
            Please note that in case of `dlt.resource` the table schema value will be overwritten and in case of `dlt.source`, the values in all resources will be overwritten.

        columns (Sequence[TColumnSchema], optional): A list of column schemas. Typed dictionary describing column names, data types, write disposition and performance hints that gives you full control over the created table schema.

        schema (Schema, optional): An explicit `Schema` object in which all table schemas will be grouped. By default `dlt` takes the schema from the source (if passed in `data` argument) or creates a default one itself.

        loader_file_format (TLoaderFileFormat, optional): The file format the loader will use to create the load package. Not all file_formats are compatible with all destinations. Defaults to the preferred file format of the selected destination.

        table_format (TTableFormat, optional): Can be "delta" or "iceberg". The table format used by the destination to store tables. Currently you can select table format on filesystem and Athena destinations.

        schema_contract (TSchemaContract, optional): On override for the schema contract settings, this will replace the schema contract settings for all tables in the schema. Defaults to None.

        refresh (TRefreshMode, optional): Fully or partially reset sources before loading new data in this run. The following refresh modes are supported:
            * `drop_sources`: Drop tables and source and resource state for all sources currently being processed in `run` or `extract` methods of the pipeline. (Note: schema history is erased)
            * `drop_resources`: Drop tables and resource state for all resources being processed. Source level state is not modified. (Note: schema history is erased)
            * `drop_data`: Wipe all data and resource state for all resources being processed. Schema is not modified.

    Raises:
        PipelineStepFailed: when a problem happened during `extract`, `normalize` or `load` steps.
    Returns:
        LoadInfo: Information on loaded data including the list of package ids and failed job statuses. Please not that `dlt` will not raise if a single job terminally fails. Such information is provided via LoadInfo.
    """
    destination = Destination.from_reference(destination)
    return pipeline().run(
        data,
        destination=destination,
        staging=staging,
        dataset_name=dataset_name,
        table_name=table_name,
        write_disposition=write_disposition,
        columns=columns,
        schema=schema,
        loader_file_format=loader_file_format,
        table_format=table_format,
        schema_contract=schema_contract,
        refresh=refresh,
    )


# plug default tracking module
from dlt.pipeline import trace, track, platform

trace.TRACKING_MODULES = [track, platform]

# setup default pipeline in the container
PipelineContext.cls__init__(pipeline)



import itertools
from collections.abc import Mapping as C_Mapping
import os
from typing import Any, Dict, ContextManager, List, Optional, Sequence, Tuple, Type, TypeVar, Union

from dlt.common import logger
from dlt.common.configuration.providers.provider import (
    ConfigProvider,
    EXPLICIT_VALUES_PROVIDER_NAME,
)
from dlt.common.configuration.const import TYPE_EXAMPLES
from dlt.common.typing import (
    AnyType,
    ConfigValueSentinel,
    StrAny,
    TSecretValue,
    get_all_types_of_class_in_union,
    is_optional_type,
    is_subclass,
    is_union_type,
)

from dlt.common.configuration.specs.base_configuration import (
    BaseConfiguration,
    CredentialsConfiguration,
    is_secret_hint,
    extract_inner_hint,
    is_context_inner_hint,
    is_base_configuration_inner_hint,
    is_valid_hint,
    is_hint_not_resolvable,
)
from dlt.common.configuration.specs import known_sections
from dlt.common.configuration.specs.config_section_context import ConfigSectionContext
from dlt.common.configuration.specs.exceptions import NativeValueError
from dlt.common.configuration.specs.pluggable_run_context import PluggableRunContext
from dlt.common.configuration.container import Container
from dlt.common.configuration.utils import log_traces, deserialize_value
from dlt.common.configuration.exceptions import (
    FieldLookupTraces,
    LookupTrace,
    ConfigFieldMissingException,
    ConfigurationWrongTypeException,
    LookupTraces,
    NestedLookupTraces,
    ValueNotSecretException,
    InvalidNativeValue,
    UnmatchedConfigHintResolversException,
)

TConfiguration = TypeVar("TConfiguration", bound=BaseConfiguration)


def resolve_configuration(
    config: TConfiguration,
    *,
    sections: Tuple[str, ...] = (),
    explicit_value: Any = None,
    accept_partial: bool = False,
) -> TConfiguration:
    if not isinstance(config, BaseConfiguration) or not hasattr(config.__class__, "__configspec__"):
        raise ConfigurationWrongTypeException(type(config))

    # try to get the native representation of the top level configuration using the config section as a key
    # allows, for example, to store connection string or service.json in their native form in single env variable or under single vault key
    # this happens only when explicit value for the configuration was not provided
    # TODO: we can move it into _resolve_configuration and also remove similar code in _resolve_config_field
    # TODO: also allow when explicit_value is dict so we can parse initial value and merge with it
    if config.__section__ and explicit_value is None:
        initial_hint = TSecretValue if isinstance(config, CredentialsConfiguration) else AnyType
        initial_value, traces = _resolve_single_value(
            config.__section__, initial_hint, AnyType, None, sections, ()
        )
        # mappings cannot be used as explicit values, we want to enumerate mappings and request the fields' values one by one
        if initial_value is not None and not isinstance(initial_value, C_Mapping):
            explicit_value = initial_value
            log_traces(None, config.__section__, type(config), initial_value, None, traces)

    return _resolve_configuration(config, sections, (), explicit_value, accept_partial)


def initialize_credentials(hint: Any, initial_value: Any) -> CredentialsConfiguration:
    """Instantiate credentials of type `hint` with `initial_value`. The initial value must be a native representation (typically string)
    or a dictionary corresponding to credential's fields. In case of union of credentials, the first configuration in the union fully resolved by
    initial value will be instantiated."""
    # use passed credentials as initial value. initial value may resolve credentials
    if is_union_type(hint):
        specs_in_union = get_all_types_of_class_in_union(hint, CredentialsConfiguration)
        assert len(specs_in_union) > 0
        first_credentials: CredentialsConfiguration = None
        for idx, spec in enumerate(specs_in_union):
            try:
                credentials = spec.from_init_value(initial_value)
                if credentials.is_resolved():
                    return credentials
                # keep first credentials in the union to return in case all of the match but not resolve
                first_credentials = first_credentials or credentials
            except (NativeValueError, NotImplementedError):
                # if none of specs in union parsed
                if idx == len(specs_in_union) - 1 and first_credentials is None:
                    raise
        return first_credentials
    else:
        assert is_subclass(hint, CredentialsConfiguration)
        return hint.from_init_value(initial_value)  # type: ignore


def inject_section(
    section_context: ConfigSectionContext, merge_existing: bool = True, lock_context: bool = False
) -> ContextManager[ConfigSectionContext]:
    """Context manager that sets section specified in `section_context` to be used during configuration resolution. Optionally merges the context already in the container with the one provided

    Args:
        section_context (ConfigSectionContext): Instance providing a pipeline name and section context
        merge_existing (bool, optional): Merges existing section context with `section_context` in the arguments by executing `merge_style` function on `section_context`. Defaults to True.
        lock_context (bool, optional): Instruct to threadlock the current thread to prevent race conditions in context injection.

    Default Merge Style:
        Gets `pipeline_name` and `sections` from existing context if they are not provided in `section_context` argument.

    Yields:
        Iterator[ConfigSectionContext]: Context manager with current section context
    """
    container = Container()
    existing_context = container[ConfigSectionContext]

    if merge_existing:
        section_context.merge(existing_context)

    return container.injectable_context(section_context, lock_context_on_yield=lock_context)


def _maybe_parse_native_value(
    config: TConfiguration, native_value: Any, embedded_sections: Tuple[str, ...]
) -> Dict[str, Any]:
    """Parses `native_value` via `config.parse_native_representation` and returns a dict of all fields that are different from
    defaults. Note that `config` will be modified.
    """
    is_explicit_instance = isinstance(native_value, BaseConfiguration)
    # if explicit value is a mapping it will be applied field by field later
    if native_value is not None and (
        not isinstance(native_value, C_Mapping) or is_explicit_instance
    ):
        try:
            try:
                # parse the native value anyway because there are configs with side effects
                config.parse_native_representation(native_value)
            except (ValueError, NotImplementedError):
                # allow native_values that are already config classes to skip parsing
                # note that we still try to do that - some like Incremental are able to initialize form those
                if not is_explicit_instance:
                    raise
            # parse native value and convert it into dict, extract the diff and use it as exact value
            # explicit_value may not be complete ie. may be a connection string without password
            # we want the resolve to still fill missing values
            native_value = {
                k: v
                for k, v in config.__class__.from_init_value(native_value)
                .as_dict_nondefault()
                .items()
            }
        except ValueError as v_err:
            raise InvalidNativeValue(type(config), type(native_value), embedded_sections, v_err)
        except NotImplementedError:
            pass

    return native_value  # type: ignore[no-any-return]


def _resolve_configuration(
    config: TConfiguration,
    explicit_sections: Tuple[str, ...],
    embedded_sections: Tuple[str, ...],
    explicit_value: Any,
    accept_partial: bool,
) -> TConfiguration:
    # do not resolve twice
    if config.is_resolved():
        return config

    config.__exception__ = None
    try:
        try:
            explicit_value = _maybe_parse_native_value(config, explicit_value, embedded_sections)
            # if native representation didn't fully resolve the config, we try to resolve field by field
            if not config.is_resolved():
                _resolve_config_fields(
                    config, explicit_value, explicit_sections, embedded_sections, accept_partial
                )
            # full configuration was resolved
            config.resolve()
        except ConfigFieldMissingException as cm_ex:
            # store the ConfigEntryMissingException to have full info on traces of missing fields
            config.__exception__ = cm_ex
            # may resolve in partial handler
            config.call_method_in_mro("on_partial")
            # if resolved then do not raise
            if not config.is_resolved() and not accept_partial:
                raise
    except Exception as ex:
        # store the exception that happened in the resolution process
        config.__exception__ = ex
        raise

    return config


def _resolve_config_fields(
    config: BaseConfiguration,
    explicit_values: StrAny,
    explicit_sections: Tuple[str, ...],
    embedded_sections: Tuple[str, ...],
    accept_partial: bool,
) -> None:
    fields = config.get_resolvable_fields()
    unresolved_fields: FieldLookupTraces = {}
    config.__resolved_fields_set__ = []

    for key, hint in fields.items():
        if key in config.__hint_resolvers__:
            # Type hint for this field is created dynamically
            hint = config.__hint_resolvers__[key](config)
        # get default and explicit values
        default_value = getattr(config, key, None)
        explicit_none = False
        explicit_value = None
        current_value = None
        # traces collected for this field
        traces: NestedLookupTraces = []

        def _set_field(is_resolvable: bool = True) -> None:
            # NOTE: we hide B023 here because the function is called only within a loop
            # collect unresolved fields
            is_resolved = config.is_field_resolved(current_value, hint)  # noqa
            if not is_resolved:
                unresolved_fields[key] = traces  # noqa
            # set value in config
            setattr(config, key, current_value)  # noqa
            # store which values were actually resolved from config providers, includes explicit values
            if (
                is_resolved
                and is_resolvable
                and (
                    default_value != current_value  # noqa
                    or explicit_value is not None  # noqa
                    or explicit_none  # noqa
                )
            ):
                config.__resolved_fields_set__.append(key)  # noqa

        if explicit_values:
            if key in explicit_values:
                # allow None to be passed in explicit values
                # so we are able to reset defaults like in regular function calls
                explicit_value = explicit_values[key]
                explicit_none = explicit_value is None
                # detect dlt.config and dlt.secrets and force injection
                if isinstance(explicit_value, ConfigValueSentinel):
                    explicit_value = None

        if is_hint_not_resolvable(hint):
            # do not resolve not resolvable, but allow for explicit values to be passed
            if not explicit_none:
                current_value = default_value if explicit_value is None else explicit_value
            traces.append(
                LookupTrace(EXPLICIT_VALUES_PROVIDER_NAME, embedded_sections, key, current_value)
            )
            _set_field(is_resolvable=False)
            continue

        # explicit none skips resolution
        if not explicit_none:
            # if hint is union of configurations, any of them must be resolved
            specs_in_union: List[Type[BaseConfiguration]] = []
            if is_union_type(hint):
                # if union contains a type of explicit value which is not a valid hint then return it
                # it could be ie. sqlalchemy Engine
                if (
                    explicit_value
                    and not is_valid_hint(type(explicit_value))
                    and get_all_types_of_class_in_union(
                        hint, type(explicit_value), with_superclass=True
                    )
                ):
                    current_value = explicit_value
                else:
                    # TODO: use default_value and explicit_value to filter the right specs from union, they constrain
                    #   base configuration
                    # if is_base_configuration_inner_hint(type(default_value)) and is_base_configuration_inner_hint(type(explicit_value)):
                    #     if type(default_value) != type(explicit_value):
                    #         raise ConfigurationValueError()
                    specs_in_union = get_all_types_of_class_in_union(hint, BaseConfiguration)
                    if len(specs_in_union) == 1:
                        is_optional = is_optional_type(hint)
                        hint = Optional[specs_in_union[0]] if is_optional else specs_in_union[0]  # type: ignore[assignment]
            if not current_value:
                if len(specs_in_union) > 1:
                    is_optional = is_optional_type(hint)
                    for idx, alt_spec in enumerate(specs_in_union):
                        # return first resolved config from an union
                        try:
                            current_value, _ = _resolve_config_field(
                                key,
                                alt_spec,
                                default_value,
                                explicit_value,
                                config,
                                config.__section__,
                                explicit_sections,
                                embedded_sections,
                                accept_partial,
                            )
                            break
                        except ConfigFieldMissingException as cfm_ex:
                            # add traces from unresolved union spec
                            traces.append(
                                LookupTraces(
                                    alt_spec.__name__,
                                    cfm_ex.config.__resolved_fields_set__,
                                    idx + 1,
                                    len(specs_in_union),
                                    cfm_ex.traces,
                                )
                            )
                        except InvalidNativeValue:
                            # if none of specs in union parsed
                            if idx == len(specs_in_union) - 1:
                                raise
                else:
                    try:
                        current_value, field_traces = _resolve_config_field(
                            key,
                            hint,
                            default_value,
                            explicit_value,
                            config,
                            config.__section__,
                            explicit_sections,
                            embedded_sections,
                            accept_partial,
                        )
                        traces.extend(field_traces)
                    except ConfigFieldMissingException as cfm_ex:
                        # if `hint` was a configuration it may not resolved
                        # collect exception traces
                        traces.append(
                            LookupTraces(
                                hint.__name__,
                                cfm_ex.config.__resolved_fields_set__,
                                0,
                                0,
                                cfm_ex.traces,
                            )
                        )
                        # keep default value
                        current_value = default_value
        else:
            # set the trace for explicit none
            traces = [LookupTrace(EXPLICIT_VALUES_PROVIDER_NAME, embedded_sections, key, None)]

        _set_field()

    # Check for dynamic hint resolvers which have no corresponding fields
    unmatched_hint_resolvers: List[str] = []
    for field_name in config.__hint_resolvers__:
        if field_name not in fields:
            unmatched_hint_resolvers.append(field_name)

    if unmatched_hint_resolvers:
        raise UnmatchedConfigHintResolversException(type(config).__name__, unmatched_hint_resolvers)

    if unresolved_fields:
        raise ConfigFieldMissingException(config, unresolved_fields)


def _resolve_config_field(
    key: str,
    hint: Type[Any],
    default_value: Any,
    explicit_value: Any,
    config: BaseConfiguration,
    config_section: str,
    explicit_sections: Tuple[str, ...],
    embedded_sections: Tuple[str, ...],
    accept_partial: bool,
) -> Tuple[Any, List[LookupTrace]]:
    inner_hint = extract_inner_hint(hint, preserve_literal=True)
    if explicit_value is not None:
        value = explicit_value
        # TODO: consider logging explicit values, currently initial values taken from configuration
        #  are passed as explicit values so that needs to be fixed first
        traces: List[LookupTrace] = [
            LookupTrace(EXPLICIT_VALUES_PROVIDER_NAME, embedded_sections, key, value)
        ]
    else:
        # resolve key value via active providers passing the original hint ie. to preserve TSecretValue
        # NOTE: if inner_hint is an embedded config, it won't be resolved and value is None
        value, traces = _resolve_single_value(
            key, hint, inner_hint, config_section, explicit_sections, embedded_sections
        )
        log_traces(config, key, hint, value, default_value, traces)
    # contexts must be resolved as a whole
    if is_context_inner_hint(inner_hint):
        pass
    # if inner_hint is BaseConfiguration then resolve it recursively
    elif is_base_configuration_inner_hint(inner_hint):
        if isinstance(explicit_value, BaseConfiguration) and explicit_value.is_resolved():
            # explicit value was resolved so use it as it is
            pass
        else:
            if default_value is not None:
                # parse default value and use it as embedded config
                if not isinstance(default_value, BaseConfiguration):
                    embedded_config = inner_hint()
                    _maybe_parse_native_value(embedded_config, default_value, embedded_sections)
                else:
                    # if default value was instance of configuration, use it as embedded initial
                    # NOTE: we do not deep copy default value. dataclasses force factories or immutable objects
                    embedded_config = default_value
            else:
                embedded_config = inner_hint()
            # only config with sections may look for initial values
            # TODO: all this code can be moved into _resolve_configuration
            # TODO: also allow when explicit_value is dict so we can parse initial value and merge with it
            if embedded_config.__section__ and explicit_value is None:
                # config section becomes the key if the key does not start with, otherwise it keeps its original value
                initial_key, initial_embedded = _apply_embedded_sections_to_config_sections(
                    embedded_config.__section__, embedded_sections + (key,)
                )
                # it must be a secret value is config is credentials
                initial_hint = (
                    TSecretValue
                    if isinstance(embedded_config, CredentialsConfiguration)
                    else AnyType
                )
                initial_value, initial_traces = _resolve_single_value(
                    initial_key, initial_hint, AnyType, None, explicit_sections, initial_embedded
                )
                if initial_value is not None and not isinstance(initial_value, C_Mapping):
                    traces.extend(initial_traces)
                    log_traces(
                        config,
                        initial_key,
                        type(embedded_config),
                        initial_value,
                        default_value,
                        initial_traces,
                    )
                    explicit_value = initial_value

            # check if hint optional
            is_optional = is_optional_type(hint)
            # accept partial becomes True if type is optional so we do not fail on optional configs that do not resolve fully
            accept_partial = accept_partial or is_optional
            # create new instance and pass value from the provider as initial, add key to sections
            # propagate top level config section, any other sections should be replaced with keys
            top_level_section = () if embedded_sections or not config_section else (config_section,)

            value = _resolve_configuration(
                embedded_config,
                explicit_sections,
                embedded_sections + top_level_section + (key,),
                explicit_value,
                accept_partial,
            )
            if value.is_partial() and is_optional:
                # do not return partially resolved optional embeds
                value = None
                default_value = None
    else:
        # if value is resolved, then deserialize and coerce it
        if value is not None:
            # do not deserialize explicit values
            if value is not explicit_value:
                value = deserialize_value(key, value, inner_hint)

    return default_value if value is None else value, traces


def _resolve_single_value(
    key: str,
    hint: Type[Any],
    inner_hint: Type[Any],
    config_section: str,
    explicit_sections: Tuple[str, ...],
    embedded_sections: Tuple[str, ...],
) -> Tuple[Optional[Any], List[LookupTrace]]:
    traces: List[LookupTrace] = []
    value = None

    container = Container()
    # get providers from container
    providers_context = container[PluggableRunContext].providers
    # we may be resolving context
    if is_context_inner_hint(inner_hint):
        # resolve context with context provider and do not look further
        value, _ = providers_context.context_provider.get_value(key, inner_hint, None)
        return value, traces
    if is_base_configuration_inner_hint(inner_hint):
        # cannot resolve configurations directly
        return value, traces

    # resolve a field of the config
    config_section, embedded_sections = _apply_embedded_sections_to_config_sections(
        config_section, embedded_sections
    )
    providers = providers_context.providers
    # get additional sections to look in from container
    sections_context = container[ConfigSectionContext]

    def look_sections(pipeline_name: str = None) -> Any:
        # start looking from the top provider with most specific set of sections first
        value: Any = None
        for provider in providers:
            if provider.is_empty:
                # do not query empty provider so they are not added to the trace
                continue

            value, provider_traces = resolve_single_provider_value(
                provider,
                key,
                hint,
                pipeline_name,
                config_section,
                # if explicit sections are provided, ignore the injected context
                explicit_sections or sections_context.sections,
                embedded_sections,
            )
            traces.extend(provider_traces)
            if value is not None:
                # value found, ignore other providers
                break

        return value

    # first try with pipeline name as section, if present
    if sections_context.pipeline_name:
        value = look_sections(sections_context.pipeline_name)
    # then without it
    if value is None:
        value = look_sections()

    return value, traces


def _build_section_lookup_paths(
    explicit_sections: Tuple[str, ...],
    embedded_sections: Tuple[str, ...],
    config_section: str,
    supports_sections: bool,
) -> List[Tuple[str, ...]]:
    """Build ordered list of section paths from most to least specific.

    Standard pop sequence removes sections from the right one at a time.
    For "sources" with section != name, also inserts (sources, name, ...embedded)
    after sources.section — so users can write sources.name.key.
    """
    if not supports_sections:
        return [()]

    ns = list(explicit_sections) + list(embedded_sections)
    base_paths: List[Tuple[str, ...]] = []

    while True:
        path = tuple(ns) + ((config_section,) if config_section else ())
        base_paths.append(path)
        if not ns:
            break
        ns.pop()

    # compact sources layout: insert sources.name path after sources.section
    if (
        len(explicit_sections) >= 3
        and explicit_sections[0] == known_sections.SOURCES
        and explicit_sections[1] != explicit_sections[2]
    ):
        compact = (explicit_sections[0], explicit_sections[2]) + tuple(embedded_sections)
        if config_section:
            compact = compact + (config_section,)
        if compact not in base_paths:
            base_paths.insert(2, compact)

    return base_paths


def resolve_single_provider_value(
    provider: ConfigProvider,
    key: str,
    hint: Type[Any],
    pipeline_name: str = None,
    config_section: str = None,
    explicit_sections: Tuple[str, ...] = (),
    embedded_sections: Tuple[str, ...] = (),
) -> Tuple[Optional[Any], List[LookupTrace]]:
    traces: List[LookupTrace] = []

    if not provider.supports_sections:
        # if provider does not support sections and pipeline name is set then ignore it
        if pipeline_name:
            return None, traces

    lookup_paths = _build_section_lookup_paths(
        explicit_sections,
        embedded_sections,
        config_section,
        provider.supports_sections,
    )

    value = None
    for path in lookup_paths:
        value, ns_key = provider.get_value(key, hint, pipeline_name, *path)
        # if secret is obtained from non secret provider, we must fail
        cant_hold_it: bool = not provider.supports_secrets and is_secret_hint(hint)
        if value is not None and cant_hold_it:
            raise ValueNotSecretException(provider.name, ns_key)

        # create trace, ignore providers that cant_hold_it
        if not cant_hold_it:
            traces.append(LookupTrace(provider.name, list(path), ns_key, value))

        if value is not None:
            break

    if value in TYPE_EXAMPLES.values():
        _emit_placeholder_warning(value, key, ns_key, provider)
    return value, traces


def _emit_placeholder_warning(
    value: Any, key: str, full_key: str, provider: ConfigProvider
) -> None:
    msg = (
        "Placeholder value encountered when resolving config or secret:\n"
        f"resolved_key: {key}, value:{value}, section: {full_key}\n"
        "Most likely, this comes from `init`-command, which creates basic templates for "
        f"non-complex configs and secrets. The provider to adjust is {provider.name}"
    )
    if bool(provider.present_locations):
        locations = "\n".join([f"\t- {os.path.abspath(loc)}" for loc in provider.present_locations])
        msg += f" at one of these locations:\n{locations}"
    logger.warning(msg=msg)


def _apply_embedded_sections_to_config_sections(
    config_section: str, embedded_sections: Tuple[str, ...]
) -> Tuple[str, Tuple[str, ...]]:
    # for the configurations that have __section__ (config_section) defined and are embedded in other configurations,
    # the innermost embedded section replaces config_section
    if embedded_sections:
        # do not add key to embedded sections if it starts with _, those sections must be ignored
        if not embedded_sections[-1].startswith("_"):
            config_section = embedded_sections[-1]
        embedded_sections = embedded_sections[:-1]

    # remove all embedded ns starting with _
    return config_section, tuple(ns for ns in embedded_sections if not ns.startswith("_"))



from typing import Dict, List, cast

from dlt.common.schema.schema import Schema
from dlt.common.storages.exceptions import SchemaNotFoundError
from dlt.common.storages.schema_storage import SchemaStorage
from dlt.common.storages.configuration import SchemaStorageConfiguration


class LiveSchemaStorage(SchemaStorage):
    def __init__(self, config: SchemaStorageConfiguration, makedirs: bool = False) -> None:
        self.live_schemas: Dict[str, Schema] = {}
        super().__init__(config, makedirs)

    def __getitem__(self, name: str) -> Schema:
        schema: Schema = None
        if name in self.live_schemas:
            schema = self.live_schemas[name]
            if not self.is_live_schema_committed(name):
                return schema
        # return new schema instance
        try:
            schema = self.load_schema(name)
        except SchemaNotFoundError:
            # a committed live schema found that is not yet written to storage
            # may happen when schema is passed explicitly via schema arg to run / pipeline
            if schema:
                return schema
            raise
        schema = self.set_live_schema(schema)
        return schema

    def save_schema(self, schema: Schema) -> str:
        # update the live schema with schema being saved, if no live schema exist, create one to be available for a getter
        schema = self.set_live_schema(schema)
        rv = super().save_schema(schema)
        return rv

    def remove_schema(self, name: str) -> None:
        super().remove_schema(name)
        # also remove the live schema
        self.live_schemas.pop(name, None)

    def commit_live_schema(self, name: str) -> str:
        """Saves live schema in storage if it was modified"""
        if not self.is_live_schema_committed(name):
            live_schema = self.live_schemas[name]
            return self._save_schema(live_schema)
        # not saved
        return None

    def is_live_schema_committed(self, name: str) -> bool:
        """Checks if live schema is present in storage and have same hash"""
        live_schema = self.live_schemas.get(name)
        if live_schema is None:
            raise SchemaNotFoundError(name, f"live-schema://{name}")
        return not live_schema.is_modified

    def set_live_schema(self, schema: Schema) -> Schema:
        """Will add or update live schema content without writing to storage."""
        live_schema = self.live_schemas.get(schema.name)
        if live_schema:
            if id(live_schema) != id(schema):
                # replace content without replacing instance
                # print(f"live schema {live_schema} updated in place")
                live_schema.replace_schema_content(schema, link_to_replaced_schema=True)
        else:
            # print(f"live schema {schema.name} created from schema")
            live_schema = self.live_schemas[schema.name] = schema
        return live_schema

    def list_schemas(self) -> List[str]:
        names = list(set(super().list_schemas()) | set(self.live_schemas.keys()))
        return names



from dlt.common.exceptions import DltException


class CliCommandInnerException(DltException):
    def __init__(self, cmd: str, msg: str, inner_exc: Exception = None) -> None:
        self.cmd = cmd
        self.inner_exc = inner_exc
        super().__init__(msg)


class CliCommandException(DltException):
    """
    Exception that can be thrown inside a cli command and can change the
    error code or docs url presented to the user. Will always be caught.
    """

    def __init__(
        self, error_code: int = -1, docs_url: str = None, raiseable_exception: Exception = None
    ) -> None:
        self.error_code = error_code
        self.docs_url = docs_url
        self.raiseable_exception = raiseable_exception


class VerifiedSourceRepoError(DltException):
    def __init__(self, msg: str, source_name: str) -> None:
        self.source_name = source_name
        super().__init__(msg)


class PipelineWasNotRun(CliCommandInnerException):
    def __init__(self, msg: str) -> None:
        super().__init__("deploy", msg, None)


class AiContextApiError(DltException):
    def __init__(self, msg: str, source_name: str = None) -> None:
        self.source_name = source_name
        super().__init__(msg)



"""Defines env variables that `dlt` uses independently of its configuration system"""

DLT_PROJECT_DIR = "DLT_PROJECT_DIR"
"""The dlt project dir is the current working directory, '.' (current working dir) by default"""

DLT_DATA_DIR = "DLT_DATA_DIR"
"""Sets default directory where pipelines' data (working directories) will be stored"""

DLT_LOCAL_DIR = "DLT_LOCAL_DIR"
"""Sets a directory where local filesystem files are stored by destinations."""

DLT_CONFIG_FOLDER = "DLT_CONFIG_FOLDER"
"""A folder (path relative to DLT_PROJECT_DIR) where config and secrets are stored"""

DLT_DEFAULT_NAMING_NAMESPACE = "DLT_DEFAULT_NAMING_NAMESPACE"
"""Python namespace default where naming modules reside, defaults to dlt.common.normalizers.naming"""

DLT_DEFAULT_NAMING_MODULE = "DLT_DEFAULT_NAMING_MODULE"
"""A module name with the default naming convention, defaults to snake_case"""

DLT_DLT_ID_LENGTH_BYTES = "DLT_DLT_ID_LENGTH_BYTES"
"""The length of the _dlt_id identifier, before base64 encoding"""

DLT_USE_JSON = "DLT_USE_JSON"
"""Type of json parser to use, defaults to orjson, may be simplejson"""

DLT_JSON_TYPED_PUA_START = "DLT_JSON_TYPED_PUA_START"
"""Start of the unicode block within the PUA used to encode types in typed json"""

DLT_PIP_TOOL = "DLT_PIP_TOOL"
"""Pip tool used to install deps in Venv"""

DLT_DISABLE_PLUGINS = "DLT_DISABLE_PLUGINS"
"""Disables detecting and instantiating plugins. Will speed up initial imports a little."""



import os
from typing import Optional, Union, Dict, List

import pyarrow as pa
import pyarrow as pa
from pyarrow import ArrowInvalid
from pyarrow import types as pat
from lancedb import DBConnection
from lancedb.common import DATA

from dlt.common import logger
from dlt.common.data_writers.escape import escape_lancedb_literal
from dlt.common.destination.exceptions import DestinationTerminalException
from dlt.common.schema import TTableSchema
from dlt.common.schema.typing import TWriteDisposition
from dlt.common.schema.utils import get_columns_names_with_prop, get_first_column_name_with_prop
from dlt.destinations.impl.lancedb.configuration import TEmbeddingProvider
from dlt.destinations.impl.lancedb.schema import add_vector_column

PROVIDER_ENVIRONMENT_VARIABLES_MAP: Dict[TEmbeddingProvider, str] = {
    "cohere": "COHERE_API_KEY",
    "gemini-text": "GOOGLE_API_KEY",
    "openai": "OPENAI_API_KEY",
    "huggingface": "HUGGINGFACE_API_KEY",
}


def set_non_standard_providers_environment_variables(
    embedding_model_provider: TEmbeddingProvider, api_key: Union[str, None]
) -> None:
    if embedding_model_provider in PROVIDER_ENVIRONMENT_VARIABLES_MAP:
        os.environ[PROVIDER_ENVIRONMENT_VARIABLES_MAP[embedding_model_provider]] = api_key or ""


def get_canonical_vector_database_doc_id_merge_key(
    load_table: TTableSchema,
) -> str:
    if merge_key := get_first_column_name_with_prop(load_table, "merge_key"):
        return merge_key
    elif primary_key := get_columns_names_with_prop(load_table, "primary_key"):
        # No merge key defined, warn and assume the first element of the primary key is `doc_id`.
        logger.warning(
            "Merge strategy selected without defined merge key - using the first element of the"
            f" primary key ({primary_key}) as merge key."
        )
        return primary_key[0]
    else:
        raise DestinationTerminalException(
            "You must specify at least a primary key in order to perform orphan removal."
        )


def create_in_filter(field_name: str, array: pa.Array) -> str:
    """Filters all rows where `field_name` is one of the values in the `array`

    If `array` is dictionary-encoded (pa.DictionaryType) we emit the
     *distinct* values stored in its dictionary.
    """
    if pat.is_dictionary(array.type):
        # use the dictionary payload (unique categorical values).
        values_py = array.dictionary.to_pylist()
    else:
        values_py = array.to_pylist()
    return f"{field_name} IN ({', '.join(map(escape_lancedb_literal, values_py))})"


def write_records(
    records: DATA,
    /,
    *,
    db_client: DBConnection,
    table_name: str,
    vector_field_name: str,
    write_disposition: Optional[TWriteDisposition] = "append",
    merge_key: Optional[str] = None,
    remove_orphans: Optional[bool] = False,
    delete_condition: Optional[str] = None,
) -> None:
    """Inserts records into a LanceDB table with automatic embedding computation.

    Args:
        records: The data to be inserted as payload.
        db_client: The LanceDB client connection.
        table_name: The name of the table to insert into.
        merge_key: Keys for update/merge operations.
        write_disposition: The write disposition - one of 'skip', 'append', 'replace', 'merge'.
        remove_orphans (bool): Whether to remove orphans after insertion or not (only merge disposition).
        filter_condition (str): If None, then all such rows will be deleted.
            Otherwise, the condition will be used as an SQL filter to limit what rows are deleted.

    Raises:
        ValueError: If the write disposition is unsupported, or `id_field_name` is not
            provided for update/merge operations.
    """
    tbl = db_client.open_table(table_name)
    tbl.checkout_latest()
    try:
        if write_disposition in ("append", "skip", "replace"):
            tbl.add(records)
        elif write_disposition == "merge":
            # LanceDB requires identical schemas for when_not_matched_by_source_delete
            # The source data schema must exactly match the target table schema (column names,
            # order, and types). Only after 22. does it work with chunks and embeddings
            target_schema = tbl.schema
            records = add_vector_column(records, target_schema, vector_field_name)
            if remove_orphans:
                tbl.merge_insert(merge_key).when_not_matched_by_source_delete(
                    delete_condition
                ).execute(records)
            else:
                tbl.merge_insert(
                    merge_key
                ).when_matched_update_all().when_not_matched_insert_all().execute(records)
        else:
            raise DestinationTerminalException(
                f"Unsupported `{write_disposition=:}` for LanceDB Destination - batch"
                " failed AND WILL **NOT** BE RETRIED."
            )
    except ArrowInvalid as e:
        raise DestinationTerminalException(
            "Python and Arrow datatype mismatch - batch failed AND WILL **NOT** BE RETRIED."
        ) from e



from typing import Iterable, Union, List, Any, Optional, cast
from itertools import chain

from dlt.common.typing import DictStrAny

from jsonpath_ng import JSONPath, Fields as JSONPathFields
from jsonpath_ng.ext import parse as _parse

TJsonPath = Union[str, JSONPath]  # Jsonpath compiled or str
TAnyJsonPath = Union[TJsonPath, Iterable[TJsonPath]]  # A single or multiple jsonpaths


def compile_path(s: TJsonPath) -> JSONPath:
    if isinstance(s, JSONPath):
        return s
    return _parse(s)


def compile_paths(s: TAnyJsonPath) -> List[JSONPath]:
    if isinstance(s, str) or not isinstance(s, Iterable):
        s = [s]
    return [compile_path(p) for p in s]


def delete_matches(paths: TAnyJsonPath, data: DictStrAny) -> None:
    """Remove all keys from `data` matching any of given json path(s).
    Filtering is done in place."""
    paths = compile_paths(paths)
    for p in paths:
        p.filter(lambda _: True, data)


def find_values(path: TJsonPath, data: DictStrAny) -> List[Any]:
    """Return a list of values found under the given json path"""
    path = compile_path(path)
    return [m.value for m in path.find(data)]


def resolve_paths(paths: TAnyJsonPath, data: DictStrAny) -> List[str]:
    """Return a list of paths resolved against `data`. The return value is a list of strings.

    Example:
    >>> resolve_paths('$.a.items[*].b', {'a': {'items': [{'b': 2}, {'b': 3}]}})
    >>> # ['a.items.[0].b', 'a.items.[1].b']
    """
    paths = compile_paths(paths)
    return list(chain.from_iterable((str(r.full_path) for r in p.find(data)) for p in paths))


def is_simple_field_path(path: JSONPath) -> bool:
    """Checks if the given path represents a simple single field name.

    Example:
    >>> is_simple_field_path(compile_path('id'))
    True
    >>> is_simple_field_path(compile_path('$.id'))
    False
    """
    return isinstance(path, JSONPathFields) and len(path.fields) == 1 and path.fields[0] != "*"


def extract_simple_field_name(path: Union[str, JSONPath]) -> Optional[str]:
    """
    Extracts a simple field name from a JSONPath if it represents a single field access.
    Returns None if the path is complex (contains wildcards, array indices, or multiple fields).

    Args:
        path: A JSONPath object or string

    Returns:
        Optional[str]: The field name if path represents a simple field access, None otherwise

    Example:
        >>> extract_simple_field_name('name')
        'name'
        >>> extract_simple_field_name('"name"')
        'name'
        >>> extract_simple_field_name('"na$me"') # Escaped characters are preserved
        'na$me'
        >>> extract_simple_field_name('"na.me"') # Escaped characters are preserved
        'na.me'
        >>> extract_simple_field_name('$.name')  # Returns None
        >>> extract_simple_field_name('$.items[*].name')  # Returns None
        >>> extract_simple_field_name('*')  # Returns None
    """
    if isinstance(path, str):
        path = compile_path(path)

    if is_simple_field_path(path):
        return cast(str, path.fields[0])

    return None


def set_value_at_path(obj: dict[str, Any], path: str, value: Any) -> None:
    """Sets a value in a nested dictionary at the specified path.

    Args:
        obj: The dictionary to modify
        path: The dot-separated path to the target location
        value: The value to set
    """
    path_parts = str(path).split(".")
    current = obj

    for part in path_parts[:-1]:
        if part not in current:
            current[part] = {}
        current = current[part]

    current[path_parts[-1]] = value



from typing import Any, List

from dlt.common.exceptions import DltException
from dlt.common.data_types import TDataType
from dlt.common.schema.typing import (
    TSchemaContractDict,
    TSchemaContractEntities,
    TSchemaEvolutionMode,
)
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.schema.typing import TColumnSchema, TColumnSchemaBase


class SchemaException(DltException):
    def __init__(self, schema_name: str, msg: str) -> None:
        self.schema_name = schema_name
        if schema_name:
            msg = f"In schema `{schema_name}`: " + msg
        super().__init__(msg)


class InvalidSchemaName(ValueError, SchemaException):
    MAXIMUM_SCHEMA_NAME_LENGTH = 64

    def __init__(self, schema_name: str) -> None:
        self.name = schema_name
        super().__init__(
            schema_name,
            f"`{schema_name}` is an invalid schema/source name. The source or schema name must be a"
            " valid Python identifier ie. a snake case function name and have maximum"
            f" {self.MAXIMUM_SCHEMA_NAME_LENGTH} characters. Ideally should contain only small"
            " letters, numbers and underscores.",
        )


# TODO: does not look like a SchemaException
# class InvalidDatasetName(ValueError, SchemaException):
#     def __init__(self, destination_name: str) -> None:
#         self.destination_name = destination_name
#         super().__init__(
#             f"Destination {destination_name} does not accept empty datasets. Please pass the"
#             " dataset name to the destination configuration ie. via dlt pipeline."
#         )


class CannotCoerceColumnException(SchemaException):
    def __init__(
        self,
        schema_name: str,
        table_name: str,
        column_name: str,
        from_type: TDataType,
        to_type: TDataType,
        coerced_value: Any,
    ) -> None:
        self.table_name = table_name
        self.column_name = column_name
        self.from_type = from_type
        self.to_type = to_type
        self.coerced_value = coerced_value
        super().__init__(
            schema_name,
            f"Cannot coerce type `{from_type}` to `{to_type}` for value `{coerced_value}` "
            f"in table `{table_name}` column `{column_name}`",
        )


class TablePropertiesConflictException(SchemaException):
    def __init__(self, schema_name: str, table_name: str, prop_name: str, val1: str, val2: str):
        self.table_name = table_name
        self.prop_name = prop_name
        self.val1 = val1
        self.val2 = val2
        super().__init__(
            schema_name,
            f"Cannot merge partial tables into table `{table_name}` due to property `{prop_name}`"
            f" with different values: `{val1} != {val2}`",
        )


class ParentTableNotFoundException(SchemaException):
    def __init__(
        self, schema_name: str, table_name: str, parent_table_name: str, explanation: str = ""
    ) -> None:
        self.table_name = table_name
        self.parent_table_name = parent_table_name
        super().__init__(
            schema_name,
            f"Parent table `{parent_table_name}` for `{table_name}` was not found in "
            f" `schema.{explanation}`",
        )


class CannotCoerceNullException(SchemaException):
    def __init__(self, schema_name: str, table_name: str, column_name: str) -> None:
        super().__init__(
            schema_name,
            f"Cannot coerce NULL in table `{table_name}` column `{column_name}` which is not"
            " nullable",
        )


class SchemaCorruptedException(SchemaException):
    pass


class SchemaIdentifierNormalizationCollision(SchemaCorruptedException):
    def __init__(
        self,
        schema_name: str,
        table_name: str,
        identifier_type: str,
        identifier_name: str,
        conflict_identifier_name: str,
        naming_name: str,
        collision_msg: str,
    ) -> None:
        if identifier_type == "column":
            table_info = f" in table `{table_name}`"
        else:
            table_info = ""
        msg = (
            f"A `{identifier_type}` name `{identifier_name}`{table_info} collides with"
            f" `{conflict_identifier_name}` after normalization with `{naming_name}` naming"
            " convention. "
            + collision_msg
        )
        self.table_name = table_name
        self.identifier_type = identifier_type
        self.identifier_name = identifier_name
        self.conflict_identifier_name = conflict_identifier_name
        self.naming_name = naming_name
        super().__init__(schema_name, msg)


class SchemaEngineNoUpgradePathException(SchemaException):
    def __init__(
        self, schema_name: str, init_engine: int, from_engine: int, to_engine: int
    ) -> None:
        self.init_engine = init_engine
        self.from_engine = from_engine
        self.to_engine = to_engine
        super().__init__(
            schema_name,
            f"No engine upgrade path in schema `{schema_name}` from engine `{init_engine}` to"
            f" `{to_engine}`, stopped at `{from_engine}`. You possibly tried to run an older dlt"
            " version against a destination you have previously loaded data to with a newer dlt"
            " version.",
        )


class DataValidationError(SchemaException):
    def __init__(
        self,
        schema_name: str,
        table_name: str,
        column_name: str,
        schema_entity: TSchemaContractEntities,
        contract_mode: TSchemaEvolutionMode,
        table_schema: Any,
        schema_contract: TSchemaContractDict,
        data_item: Any = None,
        extended_info: str = None,
    ) -> None:
        """Raised when `data_item` violates `contract_mode` on a `schema_entity` as defined by `table_schema`

        Schema, table and column names are given as a context and full `schema_contract` and causing `data_item` as an evidence.
        """
        msg = f"Table: `{table_name}` "
        if column_name:
            msg += f"Column: `{column_name}`"
        msg = (
            "In "
            + msg
            + f" . Contract on `{schema_entity}` with `{contract_mode=:}` is violated. "
            + (extended_info or "")
        )
        data_item_str = ""
        if (
            data_item
            and hasattr(data_item, "get")
            and table_schema
            and hasattr(table_schema, "get")
        ):
            # TODO: use get_columns_names_with_prop!
            identifier_columns = [
                x.get("name")
                for x in table_schema.get("columns", {}).values()
                if x.get("primary_key") or x.get("merge_key") or x.get("unique")
            ]
            if identifier_columns:
                data_item_str += "Offending data item: "
                data_item_keys = [
                    f"{column}: {data_item.get(column)}" for column in identifier_columns
                ]
                data_item_str += ", ".join(data_item_keys)
                msg += f" {data_item_str}"
        super().__init__(schema_name, msg)
        self.table_name = table_name
        self.column_name = column_name

        # violated contract
        self.schema_entity = schema_entity
        self.contract_mode = contract_mode

        # some evidence
        self.table_schema = table_schema
        self.schema_contract = schema_contract
        self.data_item = data_item


class TableNotFound(KeyError, SchemaException):
    def __init__(self, schema_name: str, table_name: str) -> None:
        self.table_name = table_name
        super().__init__(schema_name, f"Table not found: `{table_name}`")


class TableIdentifiersFrozen(SchemaException):
    def __init__(
        self,
        schema_name: str,
        table_name: str,
        to_naming: NamingConvention,
        from_naming: NamingConvention,
        details: str,
    ) -> None:
        self.table_name = table_name
        self.to_naming = to_naming
        self.from_naming = from_naming
        msg = (
            f"Attempt to normalize identifiers for a table `{table_name}` from naming"
            f" `{from_naming.name()}` to `{to_naming.name()}` changed one or more identifiers. "
        )
        msg += (
            " This table already received data and tables were created at the destination. By"
            " default changing the identifiers is not allowed. "
        )
        msg += (
            " Such changes may result in creation of a new table or a new columns while the old"
            " columns with data will still be kept. "
        )
        msg += (
            " You may disable this behavior by setting"
            " `schema.allow_identifier_change_on_table_with_data` to True or removing"
            " `x-normalizer` hints from particular tables. "
        )
        msg += f" Details: {details}"
        super().__init__(schema_name, msg)


class IncompatibleSchemaException(SchemaException):
    def __init__(self, schema_name: str, other_name: str, reason: str) -> None:
        self.other_name = other_name
        super().__init__(
            schema_name,
            f"Cannot unify with schema `{other_name}`: {reason}",
        )


class ColumnNameConflictException(SchemaException):
    pass


class UnboundColumnException(SchemaException):
    def __init__(self, schema_name: str, table_name: str, columns: List[TColumnSchemaBase]) -> None:
        self.columns = columns
        self.schema_name = schema_name
        self.table_name = table_name

        col_infos: List[str] = []
        for column in columns:
            key_type: str = ""
            if column.get("merge_key"):
                key_type = "merge key"
            elif column.get("primary_key"):
                key_type = "primary key"

            line = f"  - {column['name']}"
            if key_type or not column.get("nullable", True):
                suffix = " (marked as non-nullable"
                if key_type:
                    suffix += f" {key_type}"
                suffix += " and must have values)"
                line += suffix
            col_infos.append(line)

        msg = (
            f"The following columns in table `{table_name}` did not receive any data during"
            " this load:\n"
            + "\n".join(col_info for col_info in col_infos)
        )
        msg += (
            "\n\nThis can happen if you specify columns manually, for example, using the"
            " `merge_key`, `primary_key` or `columns` argument but they do not exist in the data.\n"
        )
        super().__init__(schema_name, msg)


class UnboundColumnWithoutTypeException(SchemaException):
    def __init__(self, schema_name: str, table_name: str, columns: List[TColumnSchemaBase]) -> None:
        self.columns = columns
        self.schema_name = schema_name
        self.table_name = table_name

        column_names = [col["name"] for col in columns]

        msg = (
            f"The following columns in table '{table_name}' did not receive any data during this"
            " load and therefore could not have their types inferred:\n"
            + "\n".join(f"  - {name}" for name in column_names)
            + "\n\nUnless type hints are provided, these columns will not be materialized in the"
            " destination.\nOne way to provide type hints is to use the 'columns' argument in"
            " the '@dlt.resource' decorator.  For"
            f" example:\n\n@dlt.resource(columns={{{repr(column_names[0])}: {{'data_type':"
            " 'text'}})\n"
        )

        super().__init__(schema_name, msg)



"""Shared TypedDicts for the dashboard."""

from typing import Union

from typing_extensions import TypedDict


class TNameValueItem(TypedDict):
    name: str
    value: str


class TPipelineListItem(TypedDict):
    name: str
    timestamp: float


class TQueryHistoryItem(TypedDict):
    query: str
    row_count: int


class _TTableListItemRequired(TypedDict):
    name: str


class TTableListItem(_TTableListItemRequired, total=False):
    """Row returned by create_table_list: required 'name' plus config-driven fields and row_count."""

    row_count: Union[int, None]


class TLoadItem(TypedDict):
    """Row from _dlt_loads after humanize_datetime_values."""

    load_id: str
    schema_name: str
    status: str
    inserted_at: str
    schema_version_hash: str



import dataclasses
from typing import Final, Optional, Any, Dict, ClassVar, List

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration
from dlt.common.typing import TSecretStrValue
from dlt.common.utils import digest128


@configspec(init=False)
class DremioCredentials(ConnectionStringCredentials):
    drivername: str = "grpc"
    username: str = None
    password: TSecretStrValue = None
    host: str = None
    port: Optional[int] = 32010
    database: str = None

    __config_gen_annotations__: ClassVar[List[str]] = ["port"]

    def to_native_credentials(self) -> str:
        from dlt.common.libs.sql_alchemy_compat import URL

        return URL.create(
            drivername=self.drivername, host=self.host, port=self.port
        ).render_as_string(hide_password=False)

    def db_kwargs(self) -> Dict[str, Any]:
        return dict(username=self.username, password=self.password)


@configspec
class DremioClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(default="dremio", init=False, repr=False, compare=False)  # type: ignore[misc]
    credentials: DremioCredentials = None
    staging_data_source: str = None
    """The name of the staging data source"""

    def fingerprint(self) -> str:
        """Returns a fingerprint of host part of a connection string"""
        if self.credentials and self.credentials.host:
            return digest128(self.credentials.host)
        return ""



import yaml
from typing import Iterator, List, Mapping, Tuple, cast

from dlt.common import logger
from dlt.common.json import json
from dlt.common.configuration import with_config
from dlt.common.configuration.accessors import config
from dlt.common.schema.utils import get_processing_hints, to_pretty_json, to_pretty_yaml
from dlt.common.storages.configuration import (
    SchemaStorageConfiguration,
    TSchemaFileFormat,
    SCHEMA_FILES_EXTENSIONS,
)
from dlt.common.storages.file_storage import FileStorage
from dlt.common.schema import Schema, verify_schema_hash
from dlt.common.typing import DictStrAny

from dlt.common.storages.exceptions import (
    InStorageSchemaModified,
    SchemaNotFoundError,
    UnexpectedSchemaName,
)


class SchemaStorage(Mapping[str, Schema]):
    SCHEMA_FILE_NAME = "schema.%s"
    NAMED_SCHEMA_FILE_PATTERN = f"%s.{SCHEMA_FILE_NAME}"

    @with_config(spec=SchemaStorageConfiguration, sections=("schema",))
    def __init__(
        self, config: SchemaStorageConfiguration = config.value, makedirs: bool = False
    ) -> None:
        self.config = config
        self.storage = FileStorage(config.schema_volume_path, makedirs=makedirs)

    def _load_schema_json(self, name: str) -> DictStrAny:
        schema_file = self._file_name_in_store(name, "json")
        return cast(DictStrAny, json.loads(self.storage.load(schema_file)))

    def load_schema(self, name: str) -> Schema:
        # loads a schema from a store holding many schemas
        storage_schema: DictStrAny = None
        try:
            storage_schema = self._load_schema_json(name)
            # prevent external modifications of schemas kept in storage
            if not verify_schema_hash(storage_schema, verifies_if_not_migrated=True):
                raise InStorageSchemaModified(name, self.config.schema_volume_path)
        except FileNotFoundError:
            # maybe we can import from external storage
            pass

        # try to import from external storage
        if self.config.import_schema_path:
            return self._maybe_import_schema(name, storage_schema)
        if storage_schema is None:
            raise SchemaNotFoundError(name, self.config.schema_volume_path)
        return Schema.from_dict(storage_schema, validate_schema=False)

    def save_schema(self, schema: Schema) -> str:
        """Saves schema to the storage and returns the path relative to storage.

        If import schema path is configured and import schema with schema.name exits, it
        will be linked to `schema` via `_imported_version_hash`. Such hash is used in `load_schema` to
        detect if import schema changed and thus to overwrite the storage schema.

        If export schema path is configured, `schema` will be exported to it.
        """
        # check if there's schema to import
        if self.config.import_schema_path:
            try:
                imported_schema = Schema.from_dict(self._load_import_schema(schema.name))
                # link schema being saved to current imported schema so it will not overwrite this save when loaded
                schema._imported_version_hash = imported_schema.stored_version_hash
            except FileNotFoundError:
                # just save the schema
                pass
        path = self._save_and_export_schema(schema)
        return path

    def save_import_schema_if_not_exists(self, schema: Schema) -> bool:
        """Saves import schema, if not exists. If schema was saved, link itself as imported from"""
        if self.config.import_schema_path:
            try:
                self._load_import_schema(schema.name)
            except FileNotFoundError:
                # save import schema only if it not exist
                import_schema_hash = self._export_schema(
                    schema, self.config.import_schema_path, remove_processing_hints=True
                )
                schema._imported_version_hash = import_schema_hash
                return True

        return False

    def remove_schema(self, name: str) -> None:
        schema_file = self._file_name_in_store(name, "json")
        self.storage.delete(schema_file)

    def has_schema(self, name: str) -> bool:
        schema_file = self._file_name_in_store(name, "json")
        return self.storage.has_file(schema_file)

    def list_schemas(self) -> List[str]:
        files = self.storage.list_folder_files(".", to_root=False)
        # extract names
        return [f.split(".")[0] for f in files]

    def clear_storage(self) -> None:
        for schema_name in self.list_schemas():
            self.remove_schema(schema_name)

    def __getitem__(self, name: str) -> Schema:
        return self.load_schema(name)

    def __len__(self) -> int:
        return len(self.list_schemas())

    def __iter__(self) -> Iterator[str]:
        for name in self.list_schemas():
            yield name

    def __contains__(self, name: str) -> bool:  # type: ignore
        return name in self.list_schemas()

    def _maybe_import_schema(self, name: str, storage_schema: DictStrAny = None) -> Schema:
        rv_schema: Schema = None
        try:
            imported_schema = self._load_import_schema(name)
            if storage_schema is None:
                # import schema when no schema in storage
                rv_schema = Schema.from_dict(imported_schema)
                # store import hash to self to track changes
                rv_schema._imported_version_hash = rv_schema.version_hash
                logger.info(
                    f"Schema {name} not present in {self.storage.storage_path} and got imported"
                    f" with version {rv_schema.stored_version} and imported hash"
                    f" {rv_schema._imported_version_hash}"
                )
                # if schema was imported, overwrite storage schema
                self._save_and_export_schema(rv_schema, check_processing_hints=True)
            else:
                # import schema when imported schema was modified from the last import
                rv_schema = Schema.from_dict(storage_schema, validate_schema=False)
                i_s = Schema.from_dict(imported_schema)
                if i_s.version_hash != rv_schema._imported_version_hash:
                    logger.warning(
                        f"Schema {name} was present in schema storage at"
                        f" {self.storage.storage_path} but will be overwritten with imported schema"
                        f" version {i_s.version} and imported hash {i_s.version_hash}"
                    )
                    tables_seen_data = rv_schema.data_tables(seen_data_only=True)
                    if tables_seen_data:
                        logger.warning(
                            f"Schema {name} in schema storage contains tables"
                            f" ({', '.join(t['name'] for t in tables_seen_data)}) that are present"
                            " in the destination. If you changed schema of those tables in import"
                            " schema, consider using one of the refresh options:"
                            " https://dlthub.com/devel/general-usage/pipeline#refresh-pipeline-data-and-state"
                        )

                    rv_schema.replace_schema_content(i_s, link_to_replaced_schema=True)
                    rv_schema._imported_version_hash = i_s.version_hash

                    # if schema was imported, overwrite storage schema
                    self._save_and_export_schema(rv_schema, check_processing_hints=True)
        except FileNotFoundError:
            # no schema to import -> skip silently and return the original
            if storage_schema is None:
                raise SchemaNotFoundError(
                    name,
                    self.config.schema_volume_path,
                    self.config.import_schema_path,
                    self.config.external_schema_format,
                )
            rv_schema = Schema.from_dict(storage_schema, validate_schema=False)

        assert rv_schema is not None
        return rv_schema

    def _load_import_schema(self, name: str) -> DictStrAny:
        import_storage = FileStorage(self.config.import_schema_path, makedirs=False)
        schema_file = self._file_name_in_store(name, self.config.external_schema_format)
        return self._parse_schema_str(
            import_storage.load(schema_file), self.config.external_schema_format
        )

    def _export_schema(
        self, schema: Schema, export_path: str, remove_processing_hints: bool = False
    ) -> str:
        if self.config.external_schema_format == "json":
            exported_schema_s = schema.to_pretty_json(
                remove_defaults=self.config.external_schema_format_remove_defaults,
                remove_processing_hints=remove_processing_hints,
            )
        elif self.config.external_schema_format == "yaml":
            exported_schema_s = schema.to_pretty_yaml(
                remove_defaults=self.config.external_schema_format_remove_defaults,
                remove_processing_hints=remove_processing_hints,
            )
        elif self.config.external_schema_format == "dbml":
            exported_schema_s = schema.to_dbml(remove_processing_hints=remove_processing_hints)
        else:
            raise ValueError(self.config.external_schema_format)

        export_storage = FileStorage(export_path, makedirs=True)
        schema_file = self._file_name_in_store(schema.name, self.config.external_schema_format)
        export_storage.save(schema_file, exported_schema_s)

        # export schema with the same settings
        stored_schema = schema.to_dict(
            remove_defaults=self.config.external_schema_format_remove_defaults,
            remove_processing_hints=remove_processing_hints,
        )
        logger.info(
            f"Schema {schema.name} exported to {export_path} with version"
            f" {stored_schema['version']}:{stored_schema['version_hash']} as"
            f" {self.config.external_schema_format}"
        )
        return stored_schema["version_hash"]

    def _save_schema(self, schema: Schema) -> str:
        """Saves schema to schema store and bumps the version"""
        schema_file = self._file_name_in_store(schema.name, "json")
        stored_schema = schema.to_dict()
        saved_path = self.storage.save(schema_file, to_pretty_json(stored_schema))
        # this should be the only place where this function is called. we bump a version and
        # clean modified status
        schema._bump_version()
        return saved_path

    def _save_and_export_schema(self, schema: Schema, check_processing_hints: bool = False) -> str:
        """Saves schema to schema store and then exports it. If the export path is the same as import
        path, processing hints will be removed.
        """
        saved_path = self._save_schema(schema)
        if self.config.export_schema_path:
            self._export_schema(
                schema,
                self.config.export_schema_path,
                self.config.export_schema_path == self.config.import_schema_path,
            )
        # if any processing hints are found we should warn the user
        if check_processing_hints:
            table_hints, _ = get_processing_hints(schema.tables)
            if table_hints:
                msg = (
                    f"Imported schema {schema.name} contains processing hints for some tables."
                    " Processing hints are used by normalizer (x-normalizer) to mark tables that"
                    " got materialized and that prevents destructive changes to the schema. In"
                    " most cases import schema should not contain processing hints because it is"
                    " mostly used to initialize tables in a new dataset. "
                )
                msg += "Affected tables are: " + ", ".join(table_hints.keys())
                logger.warning(msg)
        return saved_path

    @staticmethod
    def load_schema_file(
        path: str,
        name: str,
        extensions: Tuple[TSchemaFileFormat, ...] = SCHEMA_FILES_EXTENSIONS,
        remove_processing_hints: bool = False,
    ) -> Schema:
        storage = FileStorage(path)
        for extension in extensions:
            file = SchemaStorage._file_name_in_store(name, extension)
            if storage.has_file(file):
                parsed_schema = SchemaStorage._parse_schema_str(storage.load(file), extension)
                schema = Schema.from_dict(
                    parsed_schema, remove_processing_hints=remove_processing_hints
                )
                if schema.name != name:
                    raise UnexpectedSchemaName(name, path, schema.name)
                return schema
        raise SchemaNotFoundError(name, path)

    @staticmethod
    def _parse_schema_str(schema_str: str, extension: TSchemaFileFormat) -> DictStrAny:
        if extension == "json":
            imported_schema: DictStrAny = json.loads(schema_str)
        elif extension == "yaml":
            imported_schema = yaml.safe_load(schema_str)
        elif extension == "dbml":
            raise ValueError(extension, "Schema parser for `dbml` not yet implemented")
        elif extension == "dot":
            raise ValueError(extension, "Schema parser for `dot` not yet implemented")
        elif extension == "mermaid":
            raise ValueError(extension, "Schema parser for `mermaid` not yet implemented")
        else:
            raise ValueError(extension)
        return imported_schema

    @staticmethod
    def _file_name_in_store(name: str, fmt: TSchemaFileFormat) -> str:
        if name:
            return SchemaStorage.NAMED_SCHEMA_FILE_PATTERN % (name, fmt)
        else:
            return SchemaStorage.SCHEMA_FILE_NAME % fmt



from os.path import join
from typing import Iterable, List, Optional, Sequence

from dlt.common.data_writers.exceptions import DataWriterNotFound
from dlt.common.json import json
from dlt.common.configuration import known_sections
from dlt.common.configuration.inject import with_config
from dlt.common.destination import LOADER_FILE_FORMATS, TLoaderFileFormat
from dlt.common.configuration.accessors import config
from dlt.common.schema import TSchemaTables
from dlt.common.storages.file_storage import FileStorage
from dlt.common.storages.configuration import LoadStorageConfiguration
from dlt.common.storages.versioned_storage import VersionedStorage
from dlt.common.storages.data_item_storage import DataItemStorage
from dlt.common.storages.load_package import (
    LoadJobInfo,
    LoadPackageInfo,
    PackageStorage,
    ParsedLoadJobFileName,
    TPackageJobState,
    TLoadPackageState,
    TJobFileFormat,
)
from dlt.common.data_writers import DataWriter, FileWriterSpec, TDataItemFormat
from dlt.common.storages.exceptions import JobFileFormatUnsupported, LoadPackageNotFound


class LoadItemStorage(DataItemStorage):
    def __init__(self, package_storage: PackageStorage, writer_spec: FileWriterSpec) -> None:
        """Data item storage using `storage` to manage load packages"""
        super().__init__(writer_spec)
        self.package_storage = package_storage

    def _get_data_item_path_template(self, load_id: str, _: str, table_name: str) -> str:
        # implements DataItemStorage._get_data_item_path_template
        file_name = PackageStorage.build_job_file_name(table_name, "%s")
        file_path = self.package_storage.get_job_file_path(
            load_id, PackageStorage.NEW_JOBS_FOLDER, file_name
        )
        return self.package_storage.storage.make_full_path(file_path)


class LoadStorage(VersionedStorage):
    STORAGE_VERSION = "1.0.0"
    NORMALIZED_FOLDER = "normalized"  # folder within the volume where load packages are stored
    LOADED_FOLDER = "loaded"  # folder to keep the loads that were completely processed
    NEW_PACKAGES_FOLDER = "new"  # folder where new packages are created

    ALL_SUPPORTED_FILE_FORMATS = LOADER_FILE_FORMATS

    @with_config(spec=LoadStorageConfiguration, sections=(known_sections.LOAD,))
    def __init__(
        self,
        is_owner: bool,
        supported_file_formats: Iterable[TLoaderFileFormat],
        config: LoadStorageConfiguration = config.value,
    ) -> None:
        self.supported_loader_file_formats = supported_file_formats
        # store job file formats to add internal job formats as needed
        self.supported_job_file_formats: List[TJobFileFormat] = list(supported_file_formats)
        self.config = config
        super().__init__(
            LoadStorage.STORAGE_VERSION,
            is_owner,
            FileStorage(config.load_volume_path, "t", makedirs=is_owner),
        )
        if is_owner:
            self.initialize_storage()
        # create package storages
        self.new_packages = PackageStorage(
            FileStorage(join(config.load_volume_path, LoadStorage.NEW_PACKAGES_FOLDER)), "new"
        )
        self.normalized_packages = PackageStorage(
            FileStorage(join(config.load_volume_path, LoadStorage.NORMALIZED_FOLDER)), "normalized"
        )
        self.loaded_packages = PackageStorage(
            FileStorage(join(config.load_volume_path, LoadStorage.LOADED_FOLDER)), "loaded"
        )

    def initialize_storage(self) -> None:
        self.storage.create_folder(LoadStorage.NEW_PACKAGES_FOLDER, exists_ok=True)
        self.storage.create_folder(LoadStorage.NORMALIZED_FOLDER, exists_ok=True)
        self.storage.create_folder(LoadStorage.LOADED_FOLDER, exists_ok=True)

    def create_item_storage(self, writer_spec: FileWriterSpec) -> DataItemStorage:
        return LoadItemStorage(self.new_packages, writer_spec)

    def import_extracted_package(
        self, load_id: str, extract_package_storage: PackageStorage
    ) -> None:
        # pass the original state
        self.new_packages.create_package(
            load_id, extract_package_storage.get_load_package_state(load_id)
        )
        # import schema
        self.new_packages.save_schema(load_id, extract_package_storage.load_schema(load_id))

    def list_new_jobs(self, load_id: str) -> Sequence[str]:
        """Lists all jobs in new jobs folder of normalized package storage and checks if file formats are supported"""
        new_jobs = self.normalized_packages.list_new_jobs(load_id)
        # make sure all jobs have supported writers
        wrong_job = next(
            (
                j
                for j in new_jobs
                if ParsedLoadJobFileName.parse(j).file_format not in self.supported_job_file_formats
            ),
            None,
        )
        if wrong_job is not None:
            raise JobFileFormatUnsupported(load_id, self.supported_job_file_formats, wrong_job)
        return new_jobs

    def commit_new_load_package(self, load_id: str) -> None:
        self.storage.rename_tree(
            self.get_new_package_path(load_id), self.get_normalized_package_path(load_id)
        )

    def list_normalized_packages(self) -> Sequence[str]:
        """Lists all packages that are normalized and will be loaded or are currently loaded"""
        return self.normalized_packages.list_packages()

    def list_loaded_packages(self) -> Sequence[str]:
        """List packages that are completely loaded"""
        return self.loaded_packages.list_packages()

    def list_failed_jobs_in_loaded_package(self, load_id: str) -> Sequence[LoadJobInfo]:
        """List all failed jobs and associated error messages for a completed load package with `load_id`"""
        return self.loaded_packages.list_failed_jobs_infos(load_id)

    def begin_schema_update(self, load_id: str) -> Optional[TSchemaTables]:
        """Reads the update file from load package `load_id` and returns its content.
        Returns none if update file is already processed (deleted in commit_schema_update)
        """
        package_path = self.get_normalized_package_path(load_id)
        if not self.storage.has_folder(package_path):
            raise FileNotFoundError(package_path)
        schema_update_file = join(package_path, PackageStorage.SCHEMA_UPDATES_FILE_NAME)
        if self.storage.has_file(schema_update_file):
            schema_update: TSchemaTables = json.loads(self.storage.load(schema_update_file))
            return schema_update
        else:
            return None

    def commit_schema_update(self, load_id: str, applied_update: TSchemaTables) -> None:
        """Marks schema update as processed by removing schema update file and saving the applied
        (effective) updates as reported by destination
        """
        load_path = self.get_normalized_package_path(load_id)
        schema_update_file = join(load_path, PackageStorage.SCHEMA_UPDATES_FILE_NAME)
        applied_schema_update_file = join(
            load_path, PackageStorage.APPLIED_SCHEMA_UPDATES_FILE_NAME
        )
        # delete initial schema update
        self.storage.delete(schema_update_file)
        # save applied update
        self.storage.save(applied_schema_update_file, json.dumps(applied_update))

    def import_new_job(
        self, load_id: str, job_file_path: str, job_state: TPackageJobState = "new_jobs"
    ) -> None:
        """Adds new job by moving the `job_file_path` into `new_jobs` of package `load_id`"""
        # TODO: use normalize storage and add file type checks
        return self.normalized_packages.import_job(load_id, job_file_path, job_state)

    # def atomic_import(self, external_file_path: str, to_folder: str) -> str:
    #     """Copies or links a file at `external_file_path` into the `to_folder` effectively importing file into storage"""
    #     # LoadStorage.parse_job_file_name
    #     return self.storage.to_relative_path(
    #         FileStorage.move_atomic_to_folder(
    #             external_file_path, self.storage.make_full_path(to_folder)
    #         )
    #     )

    def complete_load_package(self, load_id: str, aborted: bool) -> None:
        self.normalized_packages.complete_loading_package(
            load_id, "aborted" if aborted else "loaded"
        )
        # move to completed
        completed_path = self.get_loaded_package_path(load_id)
        self.storage.rename_tree(self.get_normalized_package_path(load_id), completed_path)

    def maybe_remove_completed_jobs(self, load_id: str) -> None:
        """Deletes completed jobs if delete_completed_jobs config flag is set. If package has failed jobs, nothing gets deleted."""
        if self.config.delete_completed_jobs:
            self.loaded_packages.remove_completed_jobs(load_id)

    def delete_loaded_package(self, load_id: str) -> None:
        self.loaded_packages.delete_package(load_id)

    def wipe_normalized_packages(self) -> None:
        self.storage.delete_folder(self.NORMALIZED_FOLDER, recursively=True)

    def get_new_package_path(self, load_id: str) -> str:
        return join(LoadStorage.NEW_PACKAGES_FOLDER, self.new_packages.get_package_path(load_id))

    def get_normalized_package_path(self, load_id: str) -> str:
        return join(
            LoadStorage.NORMALIZED_FOLDER, self.normalized_packages.get_package_path(load_id)
        )

    def get_loaded_package_path(self, load_id: str) -> str:
        return join(LoadStorage.LOADED_FOLDER, self.loaded_packages.get_package_path(load_id))

    def get_load_package_info(self, load_id: str) -> LoadPackageInfo:
        """Gets information on normalized OR loaded package with given load_id, all jobs and their statuses."""
        try:
            return self.loaded_packages.get_load_package_info(load_id)
        except LoadPackageNotFound:
            return self.normalized_packages.get_load_package_info(load_id)

    def get_load_package_state(self, load_id: str) -> TLoadPackageState:
        """Gets state of normlized or loaded package with given load_id, all jobs and their statuses."""
        try:
            return self.loaded_packages.get_load_package_state(load_id)
        except LoadPackageNotFound:
            return self.normalized_packages.get_load_package_state(load_id)



import dataclasses
from typing import Final

from dlt.common.configuration import configspec
from dlt.common.destination import TLoaderFileFormat
from dlt.common.destination.client import (
    DestinationClientConfiguration,
    CredentialsConfiguration,
)


@configspec
class DummyClientCredentials(CredentialsConfiguration):
    def __str__(self) -> str:
        return "/dev/null"


@configspec
class DummyClientConfiguration(DestinationClientConfiguration):
    destination_type: Final[str] = dataclasses.field(default="dummy", init=False, repr=False, compare=False)  # type: ignore
    loader_file_format: TLoaderFileFormat = "jsonl"
    fail_schema_update: bool = False
    fail_prob: float = 0.0
    """probability of terminal fail"""
    retry_prob: float = 0.0
    """probability of job retry"""
    completed_prob: float = 0.0
    """probability of successful job completion"""
    exception_prob: float = 0.0
    """probability of exception transient exception when running job"""
    timeout: float = 10.0
    """timeout time"""
    fail_terminally_in_init: bool = False
    """raise terminal exception in job init"""
    fail_transiently_in_init: bool = False
    """raise transient exception in job init"""
    truncate_tables_on_staging_destination_before_load: bool = True
    """truncate tables on staging destination"""

    # new jobs workflows
    create_followup_jobs: bool = False
    """create followup job for individual jobs"""
    fail_followup_job_creation: bool = False
    """Raise generic exception during followupjob creation"""
    fail_table_chain_followup_job_creation: bool = False
    """Raise generic exception during tablechain followupjob creation"""
    create_followup_table_chain_sql_jobs: bool = False
    """create a table chain merge job which is guaranteed to fail"""
    create_followup_table_chain_reference_jobs: bool = False
    """create table chain jobs which succeed """
    credentials: DummyClientCredentials = None



from typing import Dict, Optional, cast
from dlt.common import logger
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.libs.pyarrow import get_py_arrow_datatype, pyarrow as pa

from dlt.common.schema.typing import TColumnSchema, TColumnType
from dlt.destinations.type_mapping import TypeMapperImpl

TIMESTAMP_PRECISION_TO_UNIT: Dict[int, str] = {0: "s", 3: "ms", 6: "us", 9: "ns"}
UNIT_TO_TIMESTAMP_PRECISION: Dict[str, int] = {v: k for k, v in TIMESTAMP_PRECISION_TO_UNIT.items()}


# TODO: TypeMapperImpl must be a Generic where pa.DataType will be a concrete class
class LanceDBTypeMapper(TypeMapperImpl):
    sct_to_dbt = {}

    dbt_to_sct = {
        pa.string(): "text",
        pa.float64(): "double",
        pa.bool_(): "bool",
        pa.int64(): "bigint",
        pa.binary(): "binary",
        pa.date32(): "date",
    }

    def to_destination_type(self, column: TColumnSchema, table: PreparedTableSchema) -> pa.DataType:
        # reuse existing type mapper
        dt_ = get_py_arrow_datatype(column, self.capabilities, "UTC")
        if column["data_type"] == "timestamp":
            column_name = column.get("name")
            timezone = column.get("timezone")
            precision = column.get("precision")
            if timezone is not None or precision is not None:
                logger.warning(
                    "LanceDB does not currently support column flags for timezone or precision."
                    f" These flags were used in column '{column_name}'."
                )
        return dt_

    def from_destination_type(
        self,
        db_type: pa.DataType,
        precision: Optional[int] = None,
        scale: Optional[int] = None,
    ) -> TColumnType:
        # TODO: use pyarrow helpers to convert type, this is code duplication
        if isinstance(db_type, pa.TimestampType):
            return dict(
                data_type="timestamp",
                precision=UNIT_TO_TIMESTAMP_PRECISION[db_type.unit],
                scale=scale,
            )
        if isinstance(db_type, pa.Time64Type):
            return dict(
                data_type="time",
                precision=UNIT_TO_TIMESTAMP_PRECISION[db_type.unit],
                scale=scale,
            )
        if isinstance(db_type, pa.Decimal128Type):
            precision, scale = db_type.precision, db_type.scale
            if (precision, scale) == self.capabilities.wei_precision:
                return cast(TColumnType, dict(data_type="wei"))
            return dict(data_type="decimal", precision=precision, scale=scale)
        return super().from_destination_type(db_type, precision, scale)



from __future__ import annotations

import abc
from typing import (
    TYPE_CHECKING,
    Any,
    AnyStr,
    Generator,
    Literal,
    Optional,
    Protocol,
)

from dlt.common.schema.typing import TTableSchemaColumns

if TYPE_CHECKING:
    from dlt.common.libs.pandas import DataFrame
    from dlt.common.libs.pyarrow import Table as ArrowTable

TFilterOperation = Literal["eq", "ne", "gt", "lt", "gte", "lte", "in", "not_in"]


class SupportsDataAccess(Protocol):
    """Common data access protocol shared between dbapi cursors and relations"""

    @property
    def columns_schema(self) -> TTableSchemaColumns:
        """
        Returns the expected columns schema for the result of the relation. Column types are discovered with
        sql glot query analysis and lineage. dlt hints for columns are kept in some cases. Refere to <docs-page> for more details.
        """
        ...

    def df(self, chunk_size: Optional[int] = None) -> Optional[DataFrame]:
        """Fetches the results as arrow table. Uses the native pandas implementation of the destination client cursor if available.

        Args:
            chunk_size (Optional[int]): The number of rows to fetch for this call. Defaults to None which will fetch all rows.

        Returns:
            Optional[DataFrame]: A data frame with query results.
        """
        ...

    def arrow(self, chunk_size: Optional[int] = None) -> Optional[ArrowTable]:
        """Fetches the results as arrow table. Uses the native arrow implementation of the destination client cursor if available.

        Args:
            chunk_size (Optional[int]): The number of rows to fetch for this call. Defaults to None which will fetch all rows.

        Returns:
            Optional[ArrowTable]: An arrow table with query results.
        """
        ...

    def iter_df(self, chunk_size: int) -> Generator[DataFrame, None, None]:
        """Iterates over data frames of 'chunk_size' items. Uses the native pandas implementation of the destination client cursor if available.

        Args:
            chunk_size (int): The number of rows to fetch for each iteration.

        Returns:
            Generator[DataFrame, None, None]: A generator of data frames with query results.
        """
        ...

    def iter_arrow(self, chunk_size: int) -> Generator[ArrowTable, None, None]:
        """Iterates over arrow tables of 'chunk_size' items. Uses the native arrow implementation of the destination client cursor if available.

        Args:
            chunk_size (int): The number of rows to fetch for each iteration.

        Returns:
            Generator[ArrowTable, None, None]: A generator of arrow tables with query results.
        """
        ...

    def fetchall(self) -> list[tuple[Any, ...]]:
        """Fetches all items as a list of python tuples. Uses the native dbapi fetchall implementation of the destination client cursor.

        Returns:
            list[tuple[Any, ...]]: A list of python tuples w
        """
        ...

    def fetchmany(self, chunk_size: int) -> list[tuple[Any, ...]]:
        """Fetches the first 'chunk_size' items as a list of python tuples. Uses the native dbapi fetchmany implementation of the destination client cursor.

        Args:
            chunk_size (int): The number of rows to fetch for this call.

        Returns:
            list[tuple[Any, ...]]: A list of python tuples with query results.
        """
        ...

    def iter_fetch(self, chunk_size: int) -> Generator[list[tuple[Any, ...]], Any, Any]:
        """Iterates in lists of Python tuples in 'chunk_size' chunks. Uses the native dbapi fetchmany implementation of the destination client cursor.

        Args:
            chunk_size (int): The number of rows to fetch for each iteration.

        Returns:
            Generator[list[tuple[Any, ...]], Any, Any]: A generator of lists of python tuples with query results.
        """
        ...

    def fetchone(self) -> Optional[tuple[Any, ...]]:
        """Fetches the first item as a python tuple. Uses the native dbapi fetchone implementation of the destination client cursor.

        Returns:
            Optional[tuple[Any, ...]]: A python tuple with the first item of the query results.
        """
        ...


class DBApiCursorProtocol(SupportsDataAccess, Protocol):
    """Protocol for the DBAPI cursor"""

    description: tuple[Any, ...]
    native_cursor: "DBApiCursor"
    """Cursor implementation native to current destination"""

    @property
    def columns_schema(self) -> TTableSchemaColumns: ...

    def execute(self, query: AnyStr, *args: Any, **kwargs: Any) -> None:
        """Execute a query on the cursor"""
        ...

    def close(self) -> None:
        """Close the cursor"""
        ...


class DBApiCursor(abc.ABC, DBApiCursorProtocol):
    """Protocol for the DBAPI cursor"""

    description: tuple[Any, ...]
    native_cursor: "DBApiCursor"
    """Cursor implementation native to current destination"""

    def __init__(self) -> None:
        super().__init__()
        self._columns_schema: TTableSchemaColumns = {}

    @property
    def columns_schema(self) -> TTableSchemaColumns:
        return self._columns_schema

    @columns_schema.setter
    def columns_schema(self, value: TTableSchemaColumns) -> None:
        self._columns_schema = value

    def execute(self, query: AnyStr, *args: Any, **kwargs: Any) -> None:
        """Execute a query on the cursor"""

    def close(self) -> None:
        """Close the cursor"""



from inspect import Signature, isgenerator
from typing import Any, Sequence, Set, Type

from dlt.common.exceptions import DltException
from dlt.common.reflection.exceptions import ReferenceImportError
from dlt.common.reflection.ref import ImportTrace
from dlt.common.reflection.inspect import isgeneratorfunction
from dlt.common.utils import get_callable_name


class ExtractorException(DltException):
    pass


class DltSourceException(DltException):
    pass


class DltResourceException(DltSourceException):
    def __init__(self, resource_name: str, msg: str) -> None:
        self.resource_name = resource_name
        super().__init__(msg)


class PipeException(DltException):
    def __init__(self, pipe_name: str, msg: str) -> None:
        self.pipe_name = pipe_name
        msg = f"In processing pipe `{pipe_name}`: " + msg
        super().__init__(msg)


class CreatePipeException(PipeException):
    def __init__(self, pipe_name: str, msg: str) -> None:
        super().__init__(pipe_name, msg)


class PipeItemProcessingError(PipeException):
    def __init__(self, pipe_name: str, msg: str) -> None:
        super().__init__(pipe_name, msg)


class PipeNotBoundToData(PipeException):
    def __init__(self, pipe_name: str, has_parent: bool) -> None:
        self.pipe_name = pipe_name
        self.has_parent = has_parent
        if has_parent:
            msg = (
                f"A pipe created from transformer `{pipe_name}` is unbound or its parent is unbound"
                " or empty. Provide a resource in `data_from` argument or bind resources with `|`"
                " operator."
            )
        else:
            msg = "Pipe is empty and does not have a resource at its head"
        super().__init__(pipe_name, msg)


class InvalidStepFunctionArguments(PipeException):
    def __init__(self, pipe_name: str, func_name: str, sig: Signature, call_error: str) -> None:
        self.func_name = func_name
        self.sig = sig
        super().__init__(
            pipe_name,
            f"Unable to call `{func_name}`: {call_error}. The mapping/filtering function"
            f" `{func_name}` requires first argument to take data item and optional second argument"
            f" named `meta`, but the signature is `{sig}`",
        )


class ResourceExtractionError(PipeException):
    def __init__(self, pipe_name: str, gen: Any, msg: str, kind: str) -> None:
        self.msg = msg
        self.kind = kind
        self.func_name = (
            gen.__name__
            if isgenerator(gen)
            else get_callable_name(gen) if callable(gen) else str(gen)
        )
        super().__init__(
            pipe_name,
            f"extraction of resource `{pipe_name}` in `{kind}` `{self.func_name}` caused an"
            f" exception: {msg}",
        )


class PipeGenInvalid(PipeException):
    def __init__(self, pipe_name: str, gen: Any) -> None:
        msg = (
            "A pipe generator element must be an Iterator (ie. list or generator function)."
            " Generator element is typically created from a `data` argument to pipeline.run or"
            " extract method."
        )
        msg += (
            " dlt will evaluate functions that were passed as data argument. If you passed a"
            " function the returned data type is not iterable. "
        )
        type_name = str(type(gen))
        msg += f" Generator type is {type_name}."
        if "DltSource" in type_name:
            msg += " Did you pass a @dlt.source decorated function without calling it?"
        if "DltResource" in type_name:
            msg += " Did you pass a function that returns dlt.resource without calling it?"

        super().__init__(pipe_name, msg)


class UnclosablePipe(PipeException):
    def __init__(self, pipe_name: str, gen: Any) -> None:
        type_name = str(type(gen))
        if gen_name := getattr(gen, "__name__", None):
            type_name = f"`{type_name}` ({gen_name})"
        msg = f"Pipe with gen of type `{type_name}` cannot be closed."
        if callable(gen) and isgeneratorfunction(gen):
            msg += " Closing of partially evaluated transformers is not yet supported."
        super().__init__(pipe_name, msg)


class ResourceNameMissing(DltResourceException):
    def __init__(self) -> None:
        super().__init__(
            None,
            """Resource name is missing. If you create a resource directly from data ie. from a list you must pass the name explicitly in `name` argument.
        Please note that for resources created from functions or generators, the name is the function name by default.""",
        )


class ResourceNotFoundError(DltResourceException, KeyError):
    def __init__(self, resource_name: str, context: str) -> None:
        self.resource_name = resource_name
        super().__init__(
            resource_name, f"Resource with `{resource_name}` could not be found: {context}"
        )


class InvalidResourceDataType(DltResourceException):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any], msg: str) -> None:
        self.item = item
        self._typ = _typ
        super().__init__(
            resource_name,
            f"Cannot create resource `{resource_name}` from specified data. If you want to process"
            " just one data item, enclose it in a list. "
            + msg,
        )


class InvalidParallelResourceDataType(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            "Parallel resource data must be a generator or a generator function. Resource"
            f" `{resource_name}` received data type `{_typ.__name__}`",
        )


class InvalidResourceDataTypeBasic(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            f"Resources cannot be strings or dictionaries but type `{_typ.__name__}` was provided."
            " Please pass your data in a list or as a function yielding items. If you want to"
            " process just one data item, enclose it in a list.",
        )


class InvalidResourceDataTypeMultiplePipes(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            "Resources with multiple parallel data pipes are not yet supported. This problem most"
            " often happens when you are creating a source with @dlt.source decorator that has"
            " several resources with the same name.",
        )


class InvalidTransformerDataTypeGeneratorFunctionRequired(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            "Transformer must be a function decorated with @dlt.transformer that takes data item as"
            " its first argument. Only first argument may be 'positional only'.",
        )


class InvalidTransformerGeneratorFunction(DltResourceException):
    def __init__(self, resource_name: str, func_name: str, sig: Signature, code: int) -> None:
        self.func_name = func_name
        self.sig = sig
        self.code = code
        msg = f"Transformer function `{func_name}` must take data item as its first argument. "
        if code == 1:
            msg += "The actual function does not take any arguments."
        elif code == 2:
            msg += f"Only the first argument may be 'positional only', actual signature is `{sig}`"
        elif code == 3:
            msg += f"The first argument cannot be keyword only, actual signature is `{sig}`"

        super().__init__(resource_name, msg)


class InvalidResourceDataTypeIsNone(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            "Resource data missing. Did you forget the return statement in @dlt.resource decorated"
            " function?",
        )


class InvalidResourceReturnsResource(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            "Resource returned another resource but the signature of the resource function is "
            "missing a correct type annotation (DltResource or derived class). Please annotate "
            f"function `{item}` correctly."
            "",
        )


class ResourceFunctionExpected(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            f"Expected function or callable as first parameter to resource `{resource_name}` but"
            f" type `{_typ.__name__}` found. Please decorate a function with `@dlt.resource`",
        )


class InvalidParentResourceDataType(InvalidResourceDataType):
    def __init__(self, resource_name: str, item: Any, _typ: Type[Any]) -> None:
        super().__init__(
            resource_name,
            item,
            _typ,
            f"A parent resource of `{resource_name}` is of type `{_typ.__name__}`. Did you forget"
            " to use `@dlt.resource` decorator or `resource` function?",
        )


class InvalidParentResourceIsAFunction(DltResourceException):
    def __init__(self, resource_name: str, func_name: str) -> None:
        self.func_name = func_name
        super().__init__(
            resource_name,
            f"A data source `{func_name}` of a transformer `{resource_name}` is an undecorated"
            " function. Please decorate it with `@dlt.resource` or pass to `resource` function.",
        )


class DeletingResourcesNotSupported(DltResourceException):
    def __init__(self, source_name: str, resource_name: str) -> None:
        super().__init__(
            resource_name, f"Resource cannot be removed the the source `{source_name}`"
        )


class ParametrizedResourceUnbound(DltResourceException):
    def __init__(
        self, resource_name: str, func_name: str, sig: Signature, kind: str, error: str
    ) -> None:
        self.func_name = func_name
        self.sig = sig
        msg = (
            f"The `{kind}` `{resource_name}` is parametrized and expects following arguments:"
            f" `{sig}`. Did you forget to bind the `{func_name}` function? For example from"
            f" `source.{resource_name}.bind(...)`"
        )
        if error:
            msg += f" .Details: {error}"
        super().__init__(resource_name, msg)


class ResourceNotATransformer(DltResourceException):
    def __init__(self, resource_name: str, msg: str) -> None:
        super().__init__(resource_name, f"This resource is not a transformer: {msg}")


class InconsistentTableTemplate(DltSourceException):
    def __init__(self, reason: str) -> None:
        msg = f"A set of table hints provided to the resource is inconsistent: {reason}"
        super().__init__(msg)


class DataItemRequiredForDynamicTableHints(DltResourceException):
    def __init__(self, resource_name: str) -> None:
        super().__init__(
            resource_name,
            "Resource `{resource_name}` contains a dynamic hint (i.e., a value in the"
            " `@dlt.resource` decorator is set by a function), but the data to generate the hint is"
            " missing. Make sure that each data item contains the necessary field to compute the"
            " dynamic hint value.",
        )


class SourceDataIsNone(DltSourceException):
    def __init__(self, source_name: str) -> None:
        self.source_name = source_name
        super().__init__(
            f"No data returned or yielded from source function `{source_name}`. Did you forget the"
            " return statement?"
        )


class SourceExhausted(DltSourceException):
    def __init__(self, source_name: str) -> None:
        self.source_name = source_name
        super().__init__(
            f"Source `{source_name}` is exhausted or has active iterator. You can iterate or pass"
            " the source to dlt pipeline only once."
        )


class ResourcesNotFoundError(DltSourceException):
    def __init__(
        self, source_name: str, available_resources: Set[str], requested_resources: Set[str]
    ) -> None:
        self.source_name = source_name
        self.available_resources = available_resources
        self.requested_resources = requested_resources
        self.not_found_resources = requested_resources.difference(available_resources)
        msg = (
            f"The following resources could not be found in source `{source_name}`:"
            f" {self.not_found_resources}. Available resources are: {available_resources}"
        )
        super().__init__(msg)


class SourceNotAFunction(DltSourceException):
    def __init__(self, source_name: str, item: Any, _typ: Type[Any]) -> None:
        self.source_name = source_name
        self.item = item
        self.typ = _typ
        super().__init__(
            f"First parameter to the source `{source_name}` must be a function or callable but is"
            f" type `{_typ.__name__}`. Please decorate a function with @dlt.source"
        )


class SourceIsAClassTypeError(DltSourceException):
    def __init__(self, source_name: str, _typ: Type[Any]) -> None:
        self.source_name = source_name
        self.typ = _typ
        super().__init__(
            f"First parameter to the source `{source_name}` is a class `{_typ.__name__}`. Do not"
            " decorate classes with `@dlt.source`. Instead implement `.__call__()` in your class"
            " and pass instance of such class to `dlt.source()` directly"
        )


class CurrentSourceSchemaNotAvailable(DltSourceException):
    def __init__(self) -> None:
        super().__init__(
            "Current source schema is available only when called from a function decorated with"
            " dlt.source or dlt.resource"
        )


class CurrentSourceNotAvailable(DltSourceException):
    def __init__(self) -> None:
        super().__init__(
            "Current source is available only when called from a function decorated with"
            " dlt.resource or dlt.transformer during the extract step"
        )


class ExplicitSourceNameInvalid(DltSourceException):
    def __init__(self, source_name: str, schema_name: str) -> None:
        self.source_name = source_name
        self.schema_name = schema_name
        super().__init__(
            f"Your explicit `{source_name=:}` does not match explicit `{schema_name=:}`"
        )


class UnknownSourceReference(ReferenceImportError, DltSourceException, KeyError):
    def __init__(
        self, ref: str, qualified_refs: Sequence[str], traces: Sequence[ImportTrace]
    ) -> None:
        self.ref = ref
        self.qualified_refs = qualified_refs
        super().__init__(traces=traces)

    def __str__(self) -> str:
        msg = f"`{self.ref}` is not one of already registered sources. "
        if len(self.qualified_refs) == 1 and self.qualified_refs[0] == self.ref:
            pass
        else:
            msg += (
                " Following fully qualified refs were tried in the registry:\n\t%s\n"
                % "\n\t".join(self.qualified_refs)
            )
        if self.traces:
            msg += super().__str__()
        return msg


class InvalidSourceReference(DltSourceException):
    def __init__(self, ref: str) -> None:
        self.ref = ref
        msg = f"Destination reference `{ref}` has invalid format."
        super().__init__(msg)


class IncrementalUnboundError(DltResourceException):
    def __init__(self, cursor_path: str) -> None:
        super().__init__(
            "",
            f"The incremental definition with `{cursor_path=:}` is used without being bound to the"
            " resource. This most often happens when you create dynamic resource from a generator"
            " function that uses incremental. See"
            " https://dlthub.com/docs/general-usage/incremental-loading#incremental-loading-with-last-value"
            " for an example.",
        )



from __future__ import annotations

from typing import Any, Optional, TYPE_CHECKING

from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.normalizers.naming import NamingConvention

from dlt.destinations.impl.ducklake.configuration import (
    DuckLakeClientConfiguration,
    DuckLakeCredentials,
    _get_ducklake_capabilities,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.ducklake.ducklake import DuckLakeClient


class ducklake(Destination[DuckLakeClientConfiguration, "DuckLakeClient"]):
    """Instantiate a DuckLake destination.

    A DuckLake has 3 components:
        - ducklake client: this is a `duckdb` :memory: instance with the `ducklake` ATTACHed
        - catalog: this is an SQL database storing metadata. It can be a duckdb instance
            (typically the ducklake client) or a remote database (sqlite, postgres, mysql)
        - storage: this is a filesystem where data is stored in files

    The dlt DuckLake destination gives access to the "ducklake client".
    You never have to manage the catalog and storage directly;
    this is done through the ducklake client.
    """

    def __init__(
        self,
        credentials: Optional[DuckLakeCredentials] = None,
        destination_name: Optional[str] = None,
        environment: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Configure the ducklake destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials(Optional[DuckLakeCredentials]): DuckLake credentials or instantiated connection to a DuckLake
                client (which is a duckdb instance). The DuckLake credentials include
                catalog name, catalog, and storage
            destination_name(Optional[str]): Name of a destination which. May be used as ducklake name, if
                explicit name is not set in `credentials`.
            environment (Optional[str]): Environment of the destination
            **kwargs (Any, optional): Additional arguments forwarded to the destination config
        """
        super().__init__(
            credentials=credentials,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )

    @property
    def spec(self) -> type:
        return DuckLakeClientConfiguration

    @property
    def client_class(self) -> type:
        from dlt.destinations.impl.ducklake.ducklake import DuckLakeClient

        return DuckLakeClient

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: DuckLakeClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        # disable parallel loading for duckdb and sqllite
        # NOTE: sqllite fails on concurrent catalog updates also in WAL mode
        # https://github.com/duckdb/ducklake/issues/128 does not work
        if config.credentials.catalog and config.credentials.catalog.drivername not in (
            "duckdb",
            "sqlite",
        ):
            caps.loader_parallelism_strategy = "parallel"
        return super().adjust_capabilities(caps, config, naming)

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        return _get_ducklake_capabilities()


ducklake.register()



"""Home page rendering helpers: workspace and pipeline home views."""

from typing import List, Optional

import marimo as mo

import dlt
from dlt.common.configuration.specs.pluggable_run_context import ProfilesRunContext

from dlt._workspace.helpers.dashboard import strings
from dlt._workspace.helpers.dashboard import utils
from dlt._workspace.helpers.dashboard.config import DashboardConfiguration
from dlt._workspace.helpers.dashboard.typing import TPipelineListItem
from dlt._workspace.helpers.dashboard.utils import ui


def header_controls(dlt_profile_select: mo.ui.dropdown) -> Optional[List[mo.Html]]:
    """Build profile-related header controls if profiles are enabled."""
    if isinstance(dlt.current.run_context(), ProfilesRunContext):
        return [
            dlt_profile_select,
            mo.md(
                ui.small(
                    strings.home_workspace_label.format(
                        getattr(dlt.current.run_context(), "name", None)
                    )
                )
            ),
        ]
    return None


def detect_dlt_hub() -> bool:
    """Check whether dlt.hub is available."""
    try:
        return dlt.hub.__found__
    except ImportError:
        return False


def home_header_row(
    dlt_profile_select: mo.ui.dropdown,
    dlt_pipeline_select: mo.ui.multiselect,
) -> mo.Html:
    """Shared header row with logo, profile/workspace info and pipeline select."""
    _header_controls = header_controls(dlt_profile_select)
    return mo.hstack(
        [
            mo.hstack(
                [
                    mo.image(
                        "https://dlthub.com/docs/img/dlthub-logo.png",
                        width=100,
                        alt="dltHub logo",
                    ),
                    _header_controls[0] if _header_controls else "",
                ],
                justify="start",
                gap=2,
            ),
            mo.hstack(
                [
                    _header_controls[1] if _header_controls else "",
                ],
                justify="center",
            ),
            mo.hstack(
                [
                    dlt_pipeline_select,
                ],
                justify="end",
            ),
        ],
        justify="space-between",
    )


def render_workspace_home(
    dlt_profile_select: mo.ui.dropdown,
    dlt_all_pipelines: List[TPipelineListItem],
    dlt_pipeline_select: mo.ui.multiselect,
    dlt_pipelines_dir: str,
    dlt_config: DashboardConfiguration,
) -> List[mo.Html]:
    """Render the workspace-level home view (no pipeline selected)."""
    return [
        utils.ui.section_marker(strings.app_section_name, has_content=True),
        home_header_row(dlt_profile_select, dlt_pipeline_select),
        mo.md(strings.app_title).center(),
        mo.md(strings.app_intro).center(),
        mo.callout(
            mo.vstack(
                [
                    mo.md(
                        strings.home_quick_start_title.format(
                            utils.pipeline.pipeline_link_list(dlt_config, dlt_all_pipelines)
                        )
                    ),
                    dlt_pipeline_select,
                ]
            ),
            kind="info",
        ),
        mo.md(strings.home_basics_text.format(len(dlt_all_pipelines), dlt_pipelines_dir)),
    ]


def render_pipeline_header_row(
    dlt_pipeline_name: str,
    dlt_profile_select: mo.ui.dropdown,
    dlt_pipeline_select: mo.ui.multiselect,
    buttons: List[mo.Html],
) -> List[mo.Html]:
    """Render the pipeline header row with logo, title, and action buttons."""
    header_row = home_header_row(dlt_profile_select, dlt_pipeline_select)
    pipeline_title = mo.center(
        mo.hstack(
            [
                mo.md(strings.app_title_pipeline.format(dlt_pipeline_name)),
            ],
            align="center",
        ),
    )

    return [
        mo.vstack(
            [
                mo.hstack(
                    [
                        mo.vstack(
                            [
                                header_row,
                                pipeline_title,
                            ]
                        ),
                    ],
                    justify="space-between",
                ),
            ]
        ),
        mo.hstack(buttons, justify="start"),
    ]


def render_pipeline_home(
    dlt_profile_select: mo.ui.dropdown,
    dlt_pipeline: dlt.Pipeline,
    dlt_pipeline_select: mo.ui.multiselect,
    dlt_pipelines_dir: str,
    dlt_refresh_button: mo.ui.run_button,
    dlt_pipeline_name: str,
) -> List[mo.Html]:
    """Render the pipeline-level home view (pipeline selected or requested)."""
    _buttons: List[mo.Html] = [dlt_refresh_button]
    _pipeline_execution_exception: List[mo.Html] = []
    _pipeline_execution_summary: Optional[mo.Html] = None
    _last_load_packages_info: Optional[mo.Html] = None

    _buttons.append(
        mo.ui.button(
            label=ui.small(strings.home_open_working_dir_button),
            on_click=lambda _: utils.pipeline.open_local_folder(dlt_pipeline.working_dir),
        )
    )
    if local_dir := utils.pipeline.get_local_data_path(dlt_pipeline):
        _buttons.append(
            mo.ui.button(
                label=ui.small(strings.home_open_local_data_button),
                on_click=lambda _: utils.pipeline.open_local_folder(local_dir),
            )
        )

    _stack: List[mo.Html] = [
        utils.ui.section_marker(strings.home_section_name, has_content=dlt_pipeline is not None)
    ]
    _stack.extend(
        render_pipeline_header_row(
            dlt_pipeline_name, dlt_profile_select, dlt_pipeline_select, _buttons
        )
    )

    # NOTE: last_trace does not raise on broken traces
    if trace := dlt_pipeline.last_trace:
        if _pipeline_execution_summary := utils.visualization.pipeline_execution_visualization(
            trace
        ):
            _stack.append(_pipeline_execution_summary)
        if _last_load_packages_info := mo.vstack(
            [
                mo.md(ui.small(strings.view_load_packages_text)),
                utils.visualization.load_package_status_labels(trace),
            ]
        ):
            _stack.append(_last_load_packages_info)
        if _pipeline_execution_exception := utils.pipeline.exception_section(dlt_pipeline):
            _stack.extend(_pipeline_execution_exception)
    else:
        _stack.append(
            mo.callout(
                mo.md(strings.app_pipeline_no_trace.format(dlt_pipeline_name)),
                kind="info",
            )
        )

    if not dlt_pipeline and dlt_pipeline_name:
        _stack.append(
            mo.callout(
                mo.md(strings.app_pipeline_not_found.format(dlt_pipeline_name, dlt_pipelines_dir)),
                kind="warn",
            )
        )

    return _stack
